# Four basic certainties going into this:
# 1. The COS/FUV PSFs were generated with 2 micron pixels, which is an oversample of 3 in X and 12 in Y (COS pixels are 6x24 microns)
# 2. The COS/FUV PSFs were generated on a 1024x1024 grid
# 3. Pandeia requires PSFs with the same oversample in both directions, which must go evenly into their dimensions
# 4. Pandeia requires PSFs that are square in pixel-space
# Conclusion: The 1024x1024 grid must be downsampled by 4x in Y to make it properly pixel-spaced, and cropped in X to match that size
# Neither 1024 nor 256 are divisible by 3, but 1023 and 255 are (to 341 and 85). Crop, bin by 3x. Also note
#
# 4. The COS/FUV PSFs were generated at intervals of 1 Angstrom
# 5. The PSFs are all different for all of the different filters, so this needs a detector+filter combo name like the NIRCam coronagraphy bar PSFs.
# 6. Pandeia does not need 1-Angstrom specified PSFs; for each filter, 5 or 10 will do.
# Conclusion: Make a make_psf2.py-style grid that samples them fairly appropriately, and pick those PSFs out

import sys
import glob
import bisect
from astropy.io import fits
import astropy.units as u
import numpy as np
from matplotlib import pyplot as plt

# Taken from astro_spectrum.py
def _rebin(a, shape):
    """
    Re-bin a 2D array.

    Parameters
    ----------
    a : ndarray
        Array to be re-binned
    shape : list-like
        New shape after re-binning

    Returns
    -------
    ndarray
    """
    # While the array shape is defined as x,y, it's actually implemented as y,x; hence the subscripts
    # see coord.py Grid definition for more explanation
    sh = int(shape[1]), int(a.shape[0] // shape[1]), int(shape[0]), int(a.shape[1] // shape[0])
    new = a.reshape(sh).sum(-1).sum(1)
    return new

def _master_wave():
    """
    Calculate a wavelength grid for all PSFs to be sampled from.

    Keywords
    --------
    wmin: float
        the minimum wavelength for the master grid
    wmax: float
        the maximum wavelength for the master grid
    n: int
        the number of steps in the master grid

    Returns
    -------
    wvals: ndarray
        the master log-normal wavelength grid
    """

    n = 3000
    wmin = 0.05
    wmax = 34.0
    a = (n -1)/(np.log(wmax) - np.log(wmin))
    b = n - (a * np.log(wmax))

    arr = np.arange(n)+1
    masterwave = np.exp((arr - b)/a)
    return masterwave

instrument = "cosfuv"
aperture = "primary"

filenames = glob.glob("{0}/{0}*.dat".format(sys.argv[1]))
psfs = []
psfwaves = []
for filename in filenames:
    #print(filename)
    # Step 1: Load spectrum
    filtername, cenwave, dummy, wavelength = filename.split("/")[1].split("_")
    filtername = filtername.lower()
    wavelength = int(wavelength.replace("Ang.dat","").replace("W",""))

    psfs.append(wavelength)

psfs = np.sort(psfs)
masterwave = _master_wave()*10000

good = np.where((masterwave > np.min(psfs)) & (masterwave < np.max(psfs)))
masterwave = masterwave[good]
nstep = 16
wavelengths = [masterwave[x] for x in range(len(masterwave)) if x%nstep == 0]
# Make sure they include the first and last PSF we have
wavelengths.append(psfs[0])
wavelengths.append(psfs[-1])

print(wavelengths, len(wavelengths))

for wavelength in wavelengths:
    wavestr = int(np.round(wavelength))
    avgwave = wavelength*u.angstrom
    print(wavelength)
    filename = "{0}/{0}_1024PSF_W{1}Ang.dat".format(sys.argv[1],wavestr)
    psf = []
    try:
        with open(filename) as infile:
            for line in infile:
                psf.append(line.split())
            psf = np.asarray(psf,dtype=float)
    except FileNotFoundError:
        # Interpolate if wavelength doesn't exist
        splitpoint = bisect.bisect(psfwaves, wavelength)
        wave0 = psfs[splitpoint-1]
        wave1 = psfs[splitpoint]
        print(wavestr,wave0,wave1)
        psf0 = []
        filename = "{0}/{0}_1024PSF_W{1}Ang.dat".format(sys.argv[1],wave0)
        with open(filename) as infile:
            for line in infile:
                psf0.append(line.split())
            psf0 = np.asarray(psf0,dtype=float)
        psf1 = []
        filename = "{0}/{0}_1024PSF_W{1}Ang.dat".format(sys.argv[1],wave1)
        with open(filename) as infile:
            for line in infile:
                psf1.append(line.split())
            psf1 = np.asarray(psf1,dtype=float)
        psf = psf0 + (psf1 - psf0) * (wavelength - wave0) / (wave1 - wave0)

    # Step 2: Crop 4 pixels from top and botton, and 1 from the side.
    croppsf = psf[2:-2,:-1]
    #print(croppsf.shape)
    # Step 3: Rebin by 4x in Y
    shape = (1023, 255)

    binpsf = _rebin(croppsf,shape)

    # Step 4: Pad to square
    #fullpsf = np.pad(binpsf, pad_width=((384,384),(0,0)), mode="constant", constant_values=0)
    fullpsf = binpsf

    print(fullpsf.shape, np.sum(fullpsf))

    # normalize
    fullpsf = fullpsf/np.sum(fullpsf)

    fig = plt.figure()
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)
    ax1.imshow(np.log(psf))
    ax2.imshow(np.log(fullpsf))
    plt.savefig('{}_{}lp4{}c{}_{:0.4f}.png'.format(instrument, aperture, filtername, cenwave, avgwave.to_value(u.micron)))
    plt.close()

    hdu = fits.PrimaryHDU(fullpsf)
    hdulist = fits.HDUList([hdu])   
    hdulist[0].header['WAVE0'] = avgwave.to_value(u.m)
    hdulist[0].header['INSTRUME'] = instrument
    hdulist[0].header['APERTURE'] = aperture
    hdulist[0].header['NWAVES'] = 1
    hdulist[0].header['PIXSCLX'] = 0.022/3
    hdulist[0].header['PIXSCLY'] = 0.106/3
    hdulist[0].header['DIFFLMT'] = (1.22 * avgwave.to_value(u.m) / 2.4) * 180/np.pi*3600
    hdulist[0].header['PUP_THRU'] = np.max(fullpsf.data)
    hdulist[0].header['DET_SAMP'] = 3
    hdulist[0].header['COMMENT'] = "PSFs generated with CodeV optical modelling program by the COS team"
    hdulist[0].header['COMMENT'] = "For Lifetime Position 4, in March 2017"
    hdulist[0].header['COMMENT'] = "Reformatted for Pandeia Engine 2020.1002 by A. Riedel with COSconvert.py"
    hdulist.writeto('{}_{}lp4{}c{}_{:0.4f}.fits'.format(instrument, aperture, filtername, cenwave, avgwave.to_value(u.micron)), overwrite=True)

