# Import relevant libraries
from __future__ import print_function
'''
Function to test consistency of pandeia_data

Parameters
----------
instrument: string
    lowercase name of one of the instruments supported by Pandeia

This program is meant to be run from the command line, and it will check
the following things:
1.) Is every aperture used defined globally in the config.json file?
2.) Is every filter used defined globally in the config.json file?
3.) Is every disperser used defined globally in the config.json file?
4.) Are all the filters defined as valid in config_constraints (where 
    the client looks for valid options) also defined as valid in 
    mode_config (where the engine looks)
5a.) Does every filter have a defined transmission .fits file identified 
    in config.json?
5b.) Does that transmission .fits file actually exist?
5c.) Does the filter described in the transmission .fits header match the 
     name of the filter?
6a.) Does every disperser have a defined transmission .fits file identified 
    in config.json?
6b.) Does that transmission .fits file actually exist?
6c.) Does the disperser described in the transmission .fits header match 
     the name of the disperser?
7.) Does every filter defined for a given aperture have a wavelength range set?
8.) Given the wavelength range of the filters defined for the client, do the 
    PSFs for that aperture (and, for NIRCAM, filter) cover the required 
    wavelength range?
9.) Given the wavelength range of the filters defined for the engine, do the
    PSFs for that aperture (and, for NIRCAM, filter) cover the required 
    wavelength range?
10a.) For each PSF: Does the wavelength in its .fits header match the wavelength
     in its filename?
10b.) Does the aperture (+filter, for NIRCAM) in the .fits header match the 
     aperture(+filter) in its filename?
10c.) (Where applicable) Does the coronagraph mask in the .fits header match
      the aperture in its filename?

11.) If there even IS a config_constraints block in the file
12.) Prints the sum of the PSF convolution kernel to a file called <instrument>_PSFs
'''
import json
import sys
import os
import numpy as np
from astropy.io import fits


telescope = 'jwst'

def scan(elements, all_elements, data_dir, telescope, instrument, config, elementname, aperture):
    '''
    Test of data definition consistency:
    Make sure there are .fits files defined in config.json for 
     every filter/disperser.
    Also check that the .fits files actually exist on disk.
    Also check to make sure the .fits files are what their filenames 
     and config.json claim they are
    '''
    if elements != []:
        for element in elements:
            if element not in all_elements:
                if element == None:
                    break
                print('   {0:} {1:} is not defined globally!'.format(elementname,element))
            if elementname == "Disperser":
                elementtype = '{0:}-{1:}'.format(aperture, element)
                elementpath = '{0:}_{1:}_disp'.format(aperture, element)
            else:
                elementtype = element
                elementpath = element
            try:
                elementfile = config['paths'][elementpath]
                filename = "{0:}/{1:}/{2:}/{3:}".format(data_dir, telescope, instrument, elementfile)
                if os.path.isfile(filename):
                    fitshdu = fits.open(filename)
                    if fitshdu[0].header['COMPNAME'].strip().lower() != elementtype.lower():
                        print('   {0:} {1:} says it is {2:}! ({3:})'.format(elementname, elementtype.lower(),
                                                                   fitshdu[0].header['COMPNAME'].strip().lower(),filename))
                else:
                    print('   {0:} {1:} does not exist!'.format(elementname, filename))
            except KeyError:
                print("   {0:} {1:} has no definition file!".format(elementname,elementpath))


def get_constrained(aperture,mode,config,mode_elements,element):
    '''
    Test of instrument configuration consistency:
    Get the valid aperture+filter combinations defined for this 
     instrument (for the client)
    Compare those filter/disperser combinations to the mode_config 
     definition (for the engine)
    '''
    out_elements = []
    try:
        if aperture in config['config_constraints']['apertures']:
            try:
                if mode in config['config_constraints']['apertures'][aperture][element]:
                    out_elements = config['config_constraints']['apertures'][aperture][element][mode]
                elif "default" in config['config_constraints']['apertures'][aperture][element]:
                    out_elements = config['config_constraints']['apertures'][aperture][element]['default']
                    if out_elements == None:
                        out_elements = mode_elements
                else:
                    out_elements = config['config_constraints']['apertures'][aperture][element]
            except KeyError:
                pass
    except KeyError:
        print('No config_constraints block')
        out_elements = mode_elements

    for out_element in out_elements:
        if out_element not in mode_elements:
            if out_element != None:
                print("   {0:} {1:} for aperture {2:} defined in config_constraints but not mode_config".format(element,out_element,aperture))

    return out_elements


def wavetest(waveelements, config, aperture, wavesource):
    '''
    Test of wavelengths:
    Test that every filter/disperser for a given aperture has a 
     defined wavelength range
    Deliver the overall wavelength range for this aperture (for
     which PSFs must exist)
    '''
    minwave = 99999
    maxwave = 0


    if len(waveelements) != 0:
        for waveelement in waveelements:
            try:
                if config['range'][aperture][waveelement]['wmax'] > maxwave:
                    maxwave = config['range'][aperture][waveelement]['wmax']
                if config['range'][aperture][waveelement]['wmin'] < minwave:
                    minwave = config['range'][aperture][waveelement]['wmin']
            except KeyError:
                print('   Wavelengths for {0:}/{1:} are missing! ({2:})'.format(aperture, waveelement, wavesource))
    else:
        try:
            if config['range'][aperture]['wmax'] > maxwave:
                maxwave = config['range'][aperture]['wmax']
            if config['range'][aperture]['wmin'] < minwave:
                minwave = config['range'][aperture]['wmin']
        except KeyError:
            print('   Wavelengths for {0:} are missing! ({1:})'.format(aperture, wavesource))

    return minwave, maxwave

def psftest(data_dir, telescope, instrument, mode, aperturemask, aperture, minwave_client, maxwave_client, minwave_engine, maxwave_engine, outfile):
    '''
    Conducts PSF tests, in three main categories:
    Do the wavelengths span the range config.json requires for 
     this aperture?
    Are the PSFs what their filenames claim they are?
    Print the sum of the PSF kernel to a file
    '''
    minwave = np.min([minwave_client,minwave_engine])
    maxwave = np.max([maxwave_client,maxwave_engine])

    print("\n{0:}".format(aperturemask))

    # Step 4: Scan the PSFs to make sure they are what they appear to be AND cover the wavelength range required AND sum to 1.
    # We now have wmin/wmax
    # Scan the PSFs folder to make sure the coverage for each aperture is what we say it is
    # Step 4a: Do they cover the wavelength range required?
    psflist = os.listdir("{0:}/{1:}/{2:}/psfs/".format(data_dir,telescope,instrument))
    psfcut = [x for x in psflist if aperturemask in x.split('_')[1].split('-')]
    maxpsf = 0
    minpsf = 99999
    for psf in psfcut:
        psfwave = float(psf.split('_')[2].replace('.fits',''))
        if psfwave > maxpsf:
            maxpsf = psfwave
        if psfwave < minpsf:
            minpsf = psfwave

        # Step 4b: Do they sum to 1?
        fitshdu = fits.open("{0:}/{1:}/{2:}/psfs/{3:}".format(data_dir,telescope,instrument,psf))
        outfile.write("{0:}: {1:7.5f} {2:7.5f} {3:7.5f} {4:} {5:}\n".format(psf,np.sum(fitshdu[0].data),np.max(fitshdu[0].data),np.sum(fitshdu[0].data[0:int(len(fitshdu[0].data)/3.0)][0:int(len(fitshdu[0].data)/3.0)]),fitshdu[0].header['PUPILOPD'],fitshdu[0].header['PUP_THRU']))

        # Step 4c: Are they what they appear to be?
        passed = True
        # it's only necessary that the aperture be IN that header keyword (see: nirspec, where the aperture keyword is a long list)
        if aperturemask.lower() not in fitshdu[0].header['aperture'].strip().lower():
            print('   PSF {0:} has aperture {1:}, not {2:}'.format(psf, fitshdu[0].header['aperture'].strip().lower(),aperturemask.lower()))
            passed = False
        psfwave_fits = "{0:.4f}".format(float(fitshdu[0].header['wavelen'])*1000000)
        if float(psfwave_fits) != float(psfwave):
            print('   PSF {0:} is for wavelength {1:}, not {2:}'.format(psf, psfwave_fits, psfwave))
            passed = False
        try:
            if fitshdu[0].header['coronmsk'].strip().lower().replace(' ','') != aperture.lower():
                passed = False
                print('   PSF {0:} has wrong coronmsk ({1:})'.format(psf, fitshdu[0].header['coronmsk'].strip()))
        except KeyError:
            pass

    print("Expected wavelength coverage: {0:}--{1:}; actual: {2:}--{3:}".format(minwave,maxwave,minpsf,maxpsf))
    if maxpsf < maxwave_client:
        print("   Wavelength coverage insufficent: Max Client={0:}, MaxPSFs={1:}".format(maxwave_client,maxpsf))
    if maxpsf < maxwave_engine:
        print("   Wavelength coverage insufficent: Max Engine={0:}, MaxPSFs={1:}".format(maxwave_engine,maxpsf))
    if minpsf > minwave_client:
        print("   Wavelength coverage insufficent: Min Client={0:}, MinPSFs={1:}".format(minwave_client,minpsf))
    if minpsf > minwave_engine:
        print("   Wavelength coverage insufficent: Min Engine={0:}, MinPSFs={1:}".format(minwave_engine,minpsf))


def sanitycheck(instrument):
    data_dir = os.environ['pandeia_refdata']
    config_file = open('{0:}/{1:}/{2:}/config.json'.format(data_dir,telescope,instrument))
    config = json.load(config_file)

    # Collect apertures from the config.json file
    all_apertures = config['apertures']

    # Collect filters from the config.json file
    all_filters = config['filters']

    # Collect dispersive elements from the config.json file
    all_dispersers = config['dispersers']

    outfile = open('{0:}_PSFs'.format(instrument),'w')

    for mode in config['mode_config']: # this is where the configurations for each mode are actually defined

        # Step 1: Get the valid apertures, filters, and dispersers for this instrument mode
        try:
            apertures = config['mode_config'][mode]['apertures']
        except KeyError:
            print('   No aperture definition block for {0:}'.format(mode))
            apertures = all_apertures

        try:
            mode_filters = config['mode_config'][mode]['filters']
        except KeyError:
            print('   No filter definition block for {0:}'.format(mode))
            mode_filters = all_filters

        try:
            mode_dispersers = config['mode_config'][mode]['dispersers']
        except KeyError:
            print('   No disperser definition block for {0:}'.format(mode))
            mode_dispersers = all_dispersers

        # The None filter is a no-op
        try:
            mode_filters.remove(None)
        except ValueError:
            pass

        #print(mode_filters)
        for aperture in apertures:
            print("\n{0:}/{1:}".format(mode,aperture))
            if aperture not in all_apertures:
                print('   Aperture {0:} is not defined globally!'.format(aperture))

            scan(mode_filters, all_filters, data_dir, telescope, instrument, config, 'Filter', '')
            scan(mode_dispersers, all_dispersers, data_dir, telescope, instrument, config, 'Disperser', aperture)

            # Now get the client-allowed values for this mode and aperture
            filters = get_constrained(aperture,mode,config,mode_filters,"filters")
            dispersers = get_constrained(aperture,mode,config,mode_dispersers,"dispersers")

            # The None filter is a no-op
            try:
                filters.remove(None)
            except ValueError:
                pass

            scan(filters, all_filters, data_dir, telescope, instrument, config, 'Filter', aperture)
            scan(dispersers, all_dispersers, data_dir, telescope, instrument, config, 'Disperser', aperture)

            # Step 3: Check that wavelengths are all specified
            # And now, add all of the wavelength elements and dispersers to the list
            waveelements_client = filters+dispersers
            waveelements_engine = mode_filters+mode_dispersers

            waveelements = list(set(waveelements_client+waveelements_engine))

            # The None filter is a no-op
            try:
                waveelements_client.remove(None)
            except ValueError:
                pass
            try:
                list(waveelements_engine).remove(None)
            except ValueError:
                pass

            minwave_client,maxwave_client = wavetest(waveelements_client, config, aperture, "Client")
            minwave_engine,maxwave_engine = wavetest(waveelements_engine, config, aperture, "Engine")

            if aperture in ['maskswb','masklwb']:
                for element in waveelements:  # go through all the PSFs once
                    try:
                        wmin = config['range'][aperture][element]['wmin']
                        wmax = config['range'][aperture][element]['wmax']
                        psftest(data_dir, telescope, instrument, mode, '{0:}{1:}'.format(aperture,element), aperture, wmin, wmax, wmin, wmax ,outfile)
                    except KeyError:  # most likely failure: the element has no defined wavelength range. This is checked elsewhere.
                        pass
            else:
                psftest(data_dir, telescope, instrument, mode, aperture, aperture, minwave_client, maxwave_client, minwave_engine, maxwave_engine,outfile)

    outfile.close()

if __name__ == "__main__":

    instrument = sys.argv[1]
    sanitycheck(instrument)
