import bisect
import os
import sys
import ast
import six
from collections import OrderedDict

import matplotlib.pyplot as plt
from astropy.io import fits
from pandeia.engine.psf_library import PSFLibrary
import numpy as np
from numpy import nan, inf, NaN, Inf
import scipy.interpolate as ip
import scipy

def radial_profile(image=None, ext=0, ee=False, center=None, stddev=False, pixelscalex=None, pixelscaley=None, maxradius=None,
                   normalize='None', pa_range=None, slice=0):
    """ Compute a radial profile of the image.
    This computes a discrete radial profile evaluated on the provided binsize. For a version
    interpolated onto a continuous curve, see measure_radial().
    Code taken pretty much directly from pydatatut.pdf
    Parameters
    ----------
    hdulist_or_filename : string
        FITS HDUList object or path to a FITS file.
        NaN values in the FITS data array are treated as masked and ignored in computing bin statistics.
    ext : int
        Extension in FITS file
    ee : bool
        Also return encircled energy (EE) curve in addition to radial profile?
    center : tuple of floats
        Coordinates (x,y) of PSF center, in pixel units. Default is image center.
    binsize : float
        size of step for profile. Default is pixel size.
    stddev : bool
        Compute standard deviation in each radial bin, not average?
    normalize : string
        set to 'peak' to normalize peak intensity =1, or to 'total' to normalize total flux=1.
        Default is no normalization (i.e. retain whatever normalization was used in computing the PSF itself)
    pa_range : list of floats, optional
        Optional specification for [min, max] position angles to be included in the radial profile.
        I.e. calculate that profile only for some wedge, not the full image. Specify the PA in degrees
        counterclockwise from +Y axis=0. Note that you can specify ranges across zero using negative numbers,
        such as pa_range=[-10,10].  The allowed PA range runs from -180 to 180 degrees.
    slice: integer, optional
        Slice into a datacube, for use on cubes computed by calc_datacube. Default 0 if a
        cube is provided with no slice specified.
    Returns
    --------
    results : tuple
        Tuple containing (radius, profile) or (radius, profile, EE) depending on what is requested.
        The radius gives the center radius of each bin, while the EE is given inside the whole bin
        so you should use (radius+binsize/2) for the radius of the EE curve if you want to be
        as precise as possible.
    """

    if normalize.lower() == 'peak':
        print("Calculating profile with PSF normalized to peak = 1")
        image /= image.max()
    elif normalize.lower() == 'total':
        print("Calculating profile with PSF normalized to total = 1")
        image /= image.sum()

    y, x = np.indices(image.shape, dtype=float)
    if center is None:
        # get exact center of image
        # center = (image.shape[1]/2, image.shape[0]/2)
        center = tuple((a - 1) / 2.0 for a in image.shape[::-1])

    x -= center[0]
    y -= center[1]

    binsize = np.mean([pixelscalex,pixelscaley])

    r = np.sqrt((x*pixelscalex) ** 2 + (y*pixelscaley) ** 2) / binsize  # radius in bin size steps

    if pa_range is None:
        # Use full image
        ind = np.argsort(r.flat)
        sr = r.flat[ind]  # sorted r
        sim = image.flat[ind]  # sorted image

    else:
        # Apply the PA range restriction
        pa = np.rad2deg(np.arctan2(-x, y))  # Note the (-x,y) convention is needed for astronomical PA convention
        mask = (pa >= pa_range[0]) & (pa <= pa_range[1])
        ind = np.argsort(r[mask].flat)
        sr = r[mask].flat[ind]
        sim = image[mask].flat[ind]

    ri = sr.astype(int)  # sorted r as int
    deltar = ri[1:] - ri[:-1]  # assume all radii represented (more work if not)
    rind = np.where(deltar)[0]
    nr = rind[1:] - rind[:-1]  # number in radius bin
    csim = np.nan_to_num(sim).cumsum(dtype=float)  # cumulative sum to figure out sums for each bin
    # np.nancumsum is implemented in >1.12
    tbin = csim[rind[1:]] - csim[rind[:-1]]  # sum for image values in radius bins
    radialprofile = tbin / nr 

    # pre-pend the initial element that the above code misses.
    radialprofile2 = np.empty(len(radialprofile) + 1)
    if rind[0] != 0:
        radialprofile2[0] = csim[rind[0]] / (
                rind[0] + 1)  # if there are multiple elements in the center bin, average them
    else:
        radialprofile2[0] = csim[0]  # otherwise if there's just one then just take it.
    radialprofile2[1:] = radialprofile
    rr = np.arange(
        len(radialprofile2)) * binsize + binsize * 0.5  # these should be centered in the bins, so add a half.
    if pa_range is not None:
        # for PA ranges < 45 deg or so, the innermost pixel that's valid in the mask may be
        # more than a pixel from the center. Therefore we have to include that offset here
        rr += binsize * np.floor(sr[0])


    if maxradius is not None:
        crop = rr < maxradius
        rr = rr[crop]
        radialprofile2 = radialprofile2[crop]

    if stddev:
        stddevs = np.zeros_like(radialprofile2)
        r_pix = r * binsize
        for i, radius in enumerate(rr):
            if i == 0:
                wg = np.where(r < radius + binsize / 2)
            else:
                wg = np.where((r_pix >= (radius - binsize / 2)) & (r_pix < (radius + binsize / 2)))
                # wg = np.where( (r >= rr[i-1]) &  (r <rr[i] )))
            stddevs[i] = np.nanstd(image[wg])
        return rr, stddevs

    if not ee:
        return rr, radialprofile2
    else:
        ee = csim[rind]
        return rr, radialprofile2, ee


def measure_ee(image=None, ext=0, center=None, pixelscalex=None, pixelscaley=None, normalize='None'):
    """ measure encircled energy vs radius and return as an interpolator
    Returns a function object which when called returns the Encircled Energy inside a given radius,
    for any arbitrary desired radius smaller than the image size.
    Parameters
    ----------
    HDUlist_or_filename : string
        Either a fits.HDUList object or a filename of a FITS file on disk
    ext : int
        Extension in that FITS file
    center : tuple of floats
        Coordinates (x,y) of PSF center. Default is image center.
    binsize:
        size of step for profile. Default is pixel size.
    normalize : string
        set to 'peak' to normalize peak intensity =1, or to 'total' to normalize total flux=1.
        Default is no normalization (i.e. retain whatever normalization was used in computing the PSF itself)
    Returns
    --------
    encircled_energy: function
        A function which will return the encircled energy interpolated to any desired radius.
    Examples
    --------
    >>> ee = measure_ee("someimage.fits")
    >>> print "The EE at 0.5 arcsec is ", ee(0.5)
    """

    rr, radialprofile2, ee = radial_profile(image, ext, ee=True, center=center, pixelscalex=pixelscalex, pixelscaley=pixelscaley,
                                            normalize=normalize)

    # append the zero at the center
    rr_ee = rr + (rr[1] - rr[0]) / 2.0  # add half a binsize to this, because the ee is measured inside the
    # outer edge of each annulus.
    rr0 = np.concatenate(([0], rr_ee))
    ee0 = np.concatenate(([0], ee))

    ee_fn = scipy.interpolate.interp1d(rr0, ee0, kind='cubic', bounds_error=False)

    return ee_fn

# Taken from Python-2.7 and enhanced to support np.nan and np.inf
def safe_eval(node_or_string):
    """
    Safely evaluate an expression node or a string containing a Python
    expression.  The string or node provided may only consist of the following
    Python literal structures: strings, numbers, tuples, lists, dicts, booleans,
    and None.
    """
    _safe_names = {'None': None, 'True': True, 'False': False,
                   'nan':nan, 'NaN':np.NaN,
                   'inf':inf, 'Inf':np.Inf}
    if isinstance(node_or_string, six.string_types):
        node_or_string = ast.parse(node_or_string, mode='eval')
    if isinstance(node_or_string, ast.Expression):
        node_or_string = node_or_string.body
    def _convert(node):
        # Old code used to return instances of ast.Dict. 
        # New code sometimes retuns dicts disguised as lists 
        # of tuples (actuaally, ast.List and ast.Tuple), as 
        # the first element of the 'args' attribute of an 
        # ast.Call instace. 
        if isinstance(node, ast.Call):
            result = OrderedDict()
            if len(node.args) > 0:
                for element in node.args[0].elts:
                    name  = element.elts[0].s
                    v = element.elts[1]
                    if isinstance(v, ast.Str): 
                        value = v.s
                    elif isinstance(v, ast.Num):
                        value = v.n
                    else:
                        value = None
                    result[name] = value
            return result
        elif isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, ast.List):
            return list(map(_convert, node.elts))
        elif isinstance(node, ast.Dict):
            return OrderedDict((_convert(k), _convert(v)) for k, v
                         in zip(node.keys, node.values))
        elif isinstance(node, ast.Name):
            if node.id in _safe_names:
                return _safe_names[node.id]
        elif isinstance(node, ast.BinOp) and \
             isinstance(node.op, (ast.Add, ast.Sub)) and \
             isinstance(node.right, ast.Num) and \
             isinstance(node.right.n, complex) and \
             isinstance(node.left, ast.Num) and \
             isinstance(node.left.n, six.integer_types) and \
             isinstance(node.left.n,  float):
            left = node.left.n
            right = node.right.n
            if isinstance(node.op, ast.Add):
                return left + right
            else:
                return left - right

        # when the expression contains a negative number,
        # under python 3 the negative number will be cast 
        # as an instance of UnaryOp with first operand of
        # type USub. 
        elif isinstance(node, ast.NameConstant):
            if sys.version_info[0] >= 3:
                return node.value
        elif isinstance(node, ast.UnaryOp) and \
             isinstance(node.op, ast.USub) and \
             isinstance(node.operand, ast.Num): 
            return -(node.operand.n) 

        raise ValueError('malformed string')

    return _convert(node_or_string)


def read_dict(fname):
    '''read a python dictionary from a file that was written
    with write_dict.
    '''
    # no six for this
    if sys.version_info[0] >= 3:
        f=open(fname,'r', encoding="utf-8")
    else:
        f=open(fname,'r')
    datastr = f.read()
    f.close()
    # convert DOS file to Unix - otherwise the eval will fail
    datastr = datastr.replace('\r','')
    try :
        datadict = safe_eval(datastr)
    except Exception as e:
        print('EXCEPTION:',e)
        print('cannot eval data in file ',fname)
        raise
    return datadict


refdata_path = os.environ['pandeia_refdata']

instrument = sys.argv[1]
pandeia_name = sys.argv[2]
eetabname = sys.argv[3]


#exec(open("{0}/../pyetc/instruments/{1}/{1}_aperture_fraction.dat".format(os.environ['pandeia_refdata'],instrument)))



# load EE table
eetables = f"{refdata_path}/../pyetc/pyetc/instruments/hst/{instrument}/{instrument}_aperture_fraction.dat"
eedata = read_dict(eetables)

eetable = [entry for entry in eedata['tables'] if (entry['detector'] == eetabname) and (entry['aperture_shape'] == "circle")]
eedata = eetable[0]["data"]
eedist = eedata[0][1:]
eetab = {}
for line in eedata[1:]:
    eetab[line[0]] = line[1:]

if 'swapped' in eetable:
    if eetable[0]['swapped']:
        eetable = np.rot90(eedata['data'])
        eedist = eetable[-1][1:]
        print(eedist)
        eetab = {}
        for line in eedata[:-1]:
            print(line[0])
            eetab[line[0]] = line[1:]

eewaves = list(eetab.keys())

psfs = PSFLibrary(path=f"{refdata_path}/hst/{instrument}/psfs", psf_key=pandeia_name)
pandeia_wavelengths = psfs.get_values('wave', instrument, pandeia_name)[1]
print(pandeia_wavelengths)

wavelengths = np.sort(np.append(np.asarray(eewaves),np.asarray(pandeia_wavelengths) * 10000.))
print(wavelengths)

for wavelength in wavelengths:

    try:
        psf = psfs.get_psf( wavelength/10000., instrument, pandeia_name, source_offset=(0, 0))
    except ValueError:
        pass
    else:
        oversample = psf['upsamp']

        binsize = psf['pix_scly'] #*oversample

        radius, radialprofile,ee = radial_profile(psf['int'],ee=True, pixelscalex=psf['pix_sclx'], pixelscaley=psf['pix_scly'])#, normalize='total')

        eefunc = measure_ee(psf['int'], pixelscalex=psf['pix_sclx'], pixelscaley=psf['pix_scly'])

        splitpoint = bisect.bisect(eewaves, wavelength)
        if splitpoint == 0:
            goodee = eetab[eewaves[0]]
        elif splitpoint == len(eewaves):
            goodee = eetab[eewaves[-1]]
        else:
            wave0 = eewaves[splitpoint-1]
            wave1 = eewaves[splitpoint]
            ee0 = np.array(eetab[wave0])
            ee1 = np.array(eetab[wave1])
            goodee = ee0 + (ee1 - ee0) * (wavelength - wave0) / (wave1 - wave0)
            print(goodee)

        if eetable[0]['size_units'] == "pixels":
            radius = radius / float(psf['pix_sclx'])
            limit = 60
            plt.xlabel("pixels")
        else:
            limit = 5
            plt.xlabel("arcsec")

        plt.plot(radius,ee, label="PSF")
        plt.ylabel("cumulative")
        plt.title(f"{instrument}, {pandeia_name}, {eetabname}, {wavelength}")

        plt.plot(eedist, goodee, label="EE table")
        plt.xlim(0,np.min([plt.xlim()[1],limit]))
        #plt.xlim(0,np.min([plt.xlim()[1],2.5]))

        print(eetab)

        factor = []
        for x in range(len(eedist)):
            expected = goodee[x]
            actual = eefunc(eedist[x])
            print(eedist[x],expected, actual, expected/actual)
            factor.append(expected/actual)
        print(np.nanmean(factor), np.nanstd(factor, ddof=1))

        plt.legend()
        plt.savefig(f"{instrument}_{pandeia_name}_{eetabname}_{wavelength}.png")
        plt.close()
