#!/usr/bin/env python
#
# This tool is used to generate all (so far) of thr javascript (JS) used by the client.
# We of course want to continue to generate more and more to keep the client as
# data-driven as possible.
#
# This currently dumps 3 files: config_{telescope}.js, help_links_{telescope}.js and val_constraint_data_{telescope}.js
#

import copy, glob, json, os, re, sys, time
import pandeia.engine.io_utils as iou
from pandeia.engine.jwst import name_mapper

import val_constraint_data

# Support only Python 3.*
assert sys.version_info[0] > 2, "pandeia no longer supports Python 2"

# Enforce that we are using 3.7+ to preserve dict insertion order (see below for "insertion")
assert sys.version_info[0] > 3 or sys.version_info[1] >= 7, "This tool requires Python 3.7 or higher"

# Global constants

# Instrument and modes dict. NOTE that the mode names listed here are not in alpha order, they
# are listed in the order that they will appear in the make-a-new-calc pulldowns.
SUPPORTED_MODES = {
    'hst': {
        'acs':  [], # ['sbc_imaging'],
        'cos':  ['nuv_imaging'],
        'stis': ['ccd_imaging', 'fuvmama_imaging', 'fuvmama_spectroscopy_slit', 'nuvmama_imaging', 'nuvmama_echelle'],
        'wfc3': [], # ['ir_imaging', 'uvis_imaging']
    },
    'jwst': {
        'miri':    ['imaging', 'imaging_ts', 'coronagraphy', 'lrsslit', 'lrsslitless', 'mrs', 'mrs_ts', 'target_acq'],
        'nircam':  ['sw_imaging', 'lw_imaging', 'sw_ts', 'lw_ts', 'ssgrism', 'wfgrism', 'coronagraphy', 'target_acq'],
        'niriss':  ['imaging', 'soss', 'wfss', 'ami', 'target_acq'],
        'nirspec': ['msa', 'fixed_slit', 'bots', 'ifu', 'target_acq', 'mos_conf', 'mos_ver', 'ifu_ver'],
    }
}

# These are supported modes that we are currently working on and the basics are not yet done to the 
# level of a normal PR merge.  It is expected that *most of the time* these lists are empty - that
# this feature is only used during active development.  No matter what, any modes listed here will
# only show up in "dev".
UNFINISHED_MODES = {
    'hst': { 'cos': ['nuv_imaging'], 'stis': ['fuvmama_spectroscopy_slit'], },
    'jwst': {}
}

# Uses the config_constraints system
CONSTRAINED_MODES = {
    'hst': ['stis_fuvmama_spectroscopy_slit', 'stis_nuvmama_echelle'],
    'jwst': ['miri_coronagraphy','miri_target_acq',
             'nircam_coronagraphy','nircam_lw_ts', 'nircam_sw_ts', 'nircam_target_acq',
             'niriss_target_acq',
             'nirspec_fixed_slit', 'nirspec_target_acq']
}
CONSTRAINED_MODES_VIA_DISERSERS = {
    'hst': ['stis_fuvmama_spectroscopy_slit','stis_nuvmama_echelle'],
    'jwst': []
}

DISPERSER_MODES = {
    'hst': ['stis_fuvmama_spectroscopy_slit', 'stis_nuvmama_echelle'],
    'jwst': ['miri_lrsslit', 'miri_lrsslitless', 'miri_mrs', 'miri_mrs_ts', 'nircam_ssgrism', 'nircam_wfgrism', 'niriss_soss', 'niriss_wfss'],
}

DISPFILT_MODES = {
    'hst': [],
    'jwst': ['nirspec_bots','nirspec_fixed_slit','nirspec_ifu','nirspec_msa'],
}

MULTIORDER_MODES = {
    'hst': [],
    'jwst': ['niriss_soss'],
}

# Dictionary of all possible refwave values we might need for this (populate below)
REFWAVE_DICT = None

# Permitted batch expansion - expand by a par or pars.  The order they appear here in this list
# is the order they will appear in the client pull-down menu (per calc).
# In 1.2.2 was: miri_imaging, miri_TA, nircam_LW, nircam_SW, niriss_ami, niriss_imaging, niriss_wfss, nirspec_TA
# Syntax:
#    Each dict in the list for an instr_mode represents an expansion the user may do.  In that dict:
#        - name: the string shown on the menu pulldown (this is used as a key and must be unique)
#        - pars: the list of par names whose values will be zippered together to make a single calc (a tuple represent an all-comb)
#        - match: a way to hold one item constant in an expansion, e.g. if match is aperture, the client will allow only those calcs which match current calc's aperture
EXPANSIONS = {

  'hst': {
    'stis_ccd_imaging':     [ {'name': 'Expand over filters',                  'pars': ['filter',]}, ],
    'stis_fuvmama_imaging': [ {'name': 'Expand over filters',                  'pars': ['filter',]}, ],
    'stis_nuvmama_imaging': [ {'name': 'Expand over filters',                  'pars': ['filter',]}, ],
  },

  'jwst': {
#   'miri_coronagraphy':   [ {'name': 'Expand over mask/filter pairs', 'pars': ['aperture',]}, ],
    'miri_imaging':        [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'miri_imaging_ts':     [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'miri_mrs':            [ {'name': 'Expand Channels',                       'pars': [('aperture','disperser'),'reference_wavelength'], 'match': 'disperser'},
                             {'name': 'Expand Wavelength Ranges',              'pars': [('aperture','disperser'),'reference_wavelength'], 'match': 'aperture'},
                             {'name': 'Expand Channels and Wavelength Ranges', 'pars': [('aperture','disperser'),'reference_wavelength']}, ], # allow all 12 combos
    'miri_mrs_ts':         [ {'name': 'Expand Channels',                       'pars': [('aperture','disperser'),'reference_wavelength'], 'match': 'disperser'},
                             {'name': 'Expand Wavelength Ranges',              'pars': [('aperture','disperser'),'reference_wavelength'], 'match': 'aperture'},
                             {'name': 'Expand Channels and Wavelength Ranges', 'pars': [('aperture','disperser'),'reference_wavelength']}, ], # allow all 12 combos
    'miri_target_acq':     [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'nircam_coronagraphy': [ {'name': 'Expand over Masks',                     'pars': [('aperture','filter')], 'match': 'filter'},
                             {'name': 'Expand over Filters',                   'pars': [('aperture','filter')], 'match': 'aperture'}, ],
    'nircam_lw_imaging':   [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'nircam_sw_imaging':   [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'nircam_sw_ts':        [ {'name': 'Expand over filters',                   'pars': [('aperture','filter')], 'match': 'aperture'}, ],
    'nircam_lw_ts':        [ {'name': 'Expand over filters',                   'pars': [('aperture','filter')], 'match': 'aperture'}, ],
    'niriss_ami':          [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'niriss_imaging':      [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'niriss_wfss':         [ {'name': 'Expand over filters',                   'pars': ['filter','reference_wavelength']}, ],
    'nirspec_ifu_ver':     [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'nirspec_mos_conf':    [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'nirspec_mos_ver':     [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
    'nirspec_target_acq':  [ {'name': 'Expand over filters',                   'pars': ['filter',]}, ],
  },
}

SPEC_MODENAMES = {
    'jwst': ['bots','fixed_slit','ifu','lrsslit','lrsslitless','mrs','mrs_ts','msa','soss','ssgrism','wfgrism','wfss'],
    'hst':  ['fuvmama_spectroscopy_slit', 'nuvmama_echelle'],
}

# Strategy names permitted in the engine which are not supported in the UI
NON_UI_STRATEGIES = ['ifuapphot','tacentroid']

# reverse mapping of STRATEGY_LETTERS in /pandeia_test/test_utilities/client/client_test_lib.py
# Strategy names-to-letter mapping.  This convention only used in the UI (for now).
STRATEGY_LETTERS = {
    "imagingapphot":    "a",
    "specapphot":       "b",
    "msafullapphot":    "c",
    "coronagraphy":     "d",
    "ifunodinscene":    "e",
    "ifunodoffscene":   "f",
    "soss":             "g",
    "taphot":           "h",
    "msashutterapphot": "j",
    "msaapphot":        "k",
}

# Due to the way our data and data constraints are laid out in the config files, we might have
# aperture constraints shared by multiple instr_mode values, and this can cause the constraints
# sanity-checker to try to sanity-check combinations that aren't even possible in the UI.
# So we list here any such impossible combinations that the sanity checker might nevertheless
# try to use and then get snagged on.
# Here is one such example (user cannot select a filter by itself, with nirspec_fixed_slit).
IGNORE_CONSTRAINT_SANITY_CHECKS = [
    ('nirspec_fixed_slit', 'filters'),
    ('stis_fuvmama_spectroscopy_slit', 'slits'),
]

# Used by the time series hacks
TS_DOUBLE_FILTS = {
    'sw': ('f162m','f164n'),
    'lw': ('f323n','f405n','f466n','f470n'),
}

HST_BG_AIRGLOW_CHOICES = [['none', 'No Airglow (Visible/IR Observations)'], 'low', 'average', 'high']

HST_BG_EARTHSHINE_CHOICES = ['none', 'average', 'high', ['extremely_high', 'Extremely High']]

HST_BG_ZODIACALLIGHT_CHOICES = ['none', 'low', 'average', 'high']


def camel_case(s):
    # return camel-case version of a string s.
    # for now just handle the first letter but later expand this if needed
    if not s:
        return s
    return s[0:1].upper()+s[1:]


def strip_meta(d):
    """
    Recursively strip any meta-information (e.g. 'meta') tags from a dict, d.
    This function modifies the dict in place.
    """
    if isinstance(d, dict):
        if 'meta' in d:
            d.pop('meta')
        if '-delivered-for-version-' in d:
            d.pop('-delivered-for-version-')
        for k in d:
            strip_meta(d[k])


def hacks_for_sw_ts_submodes():
    """
    We are showing/using a submode/flavor in the client that has no reality/existence
    in the data (May 2019).  So this has to be hacked here.
    """
    # the values here are the mode names of modes to be paired with
    return [ ['lw_ts','LW Imaging Time Series'],['ssgrism','LW Grism Time Series'] ];


def hacks_for_ts_apertures(mode, aperinfo):
    """ currently (201905) the client view for "pupil" selection for the NIRCam *_ts modes
    (as clearly requested by our partners) is so far from our aperture-based data layout
    that we need to hack this a bit to show some of the double-filters as aperture choices

    !!
    UPDATE PARALLEL DATA/LOGIC IN: pandeia_test/test_utilities/client/regen_lib/vary_all_calcs.py
    !!
    """
    # checks
    assert len(aperinfo) > 0, 'hacks_for_ts_apertures: bad aperinfo'

    if mode == 'sw_ts':
        junk = [item[0] for item in aperinfo]
        assert 'sw' in junk and 'wlp4' in junk, 'hacks_for_ts_apertures: unexpected aperinfo: %s' % aperinfo

        # Turn the simpler list-of-lists aperinfo into a dict keyed by submode/flavor
        apers_for_lw_ts   = [item for item in aperinfo if item[0]     in ('sw','wlp8__ts')]
        apers_for_ssgrism = [item for item in aperinfo if item[0] not in ('sw','wlp8__ts')]
        retval = {
            'lw_ts':   apers_for_lw_ts,
            'ssgrism': apers_for_ssgrism
        }
        for filt in TS_DOUBLE_FILTS['sw']:
            retval['lw_ts'].append(['sw__ui__%s'%filt,filt.upper()])
        retval['lw_ts'] = sorted(retval['lw_ts'])
        return retval

    elif mode == 'lw_ts':
        junk = [item[0] for item in aperinfo]
        assert 'lw' in junk, 'hacks_for_ts_apertures: unexpected aperinfo: %s' % aperinfo
        retval = aperinfo
        for filt in TS_DOUBLE_FILTS['lw']:
            retval.append(['lw__ui__%s'%filt,filt.upper()])
        retval.sort()
        return retval

    else: # not used
        return aperinfo


def hacks_for_ts_dblfilts(mode, constraints):
    """ currently (201905) the client view for "pupil" selection for the NIRCam *_ts modes
    (as clearly requested by our partners) is so far from our aperture-based data layout
    that we need to hack this a bit to show some of the double-filters as aperture choices """

    # create extra "apertures" for the double-filter cases, and pull out those
    # filter values from the regular aperture="sw" list

    if mode not in ('lw_ts','sw_ts'):
        return constraints

    main_aper = mode[:2] # e.g. 'lw' or 'sw'

    # checks
    assert main_aper in constraints, 'hacks_for_ts_dblfilts: constraints dict not as expected:\n%s' % constraints
    for filt in TS_DOUBLE_FILTS[main_aper]:
        assert filt in constraints[main_aper]['filters'], 'hacks_for_ts_dblfilts: %s not found where expected' % filt

    for filt in TS_DOUBLE_FILTS[main_aper]:
        # add the new "apertures" to the constraints dict
        constraints['%s__ui__%s' % (main_aper,filt)] = {
            'filters': [filt,],
            'subarrays': constraints[main_aper]['subarrays'],
        }
        # remove those filter choices now from the main_aper dict
        constraints[main_aper]['filters'].remove(filt)

    return constraints


def hack_hst_cfg(request, instr_mode=''):
    """ put this HACK in place until the HST stuff is worked out """
    if request == 'instruments':
        return ['wfc3', 'acs', 'nicmos'] # ['acs','wfc3']
    if request == 'modes':
        return {'wfc3': ['ir', 'uvis1'], 'acs': ['wfc1'], 'nicmos': ['2']};
#       return {'acs': ['hrc','sbc','wfc1'], 'wfc3': ['uvis1','uvis2']};
    if request == 'filters':
        raise RuntimeError('We should no longer be using this code')
        if instr_mode == 'acs,hrc': return ['f220w','f330w','f658n','f850lp','f892n']
        if instr_mode == 'acs,sbc': return ['f122m','f125lp','f140lp','f165lp']
        if instr_mode == 'acs,wfc1': return ['f435w','f550m','f606w','f658n','f814w','f850lp','f892n']
        if instr_mode == 'wfc3,uvis1': return ['f225w','f275w','f300x','f336w','f350lp','f438w','f555w','f606w','f656n','f673n','f814w','f850lp']
        if instr_mode == 'wfc3,uvis2': return ['f225w','f275w','f300x','f336w','f350lp','f438w','f555w','f606w','f656n','f673n','f814w','f850lp']
    raise RuntimeError('how did we get here?')


def special_sort(keys):
    """
    Sort of a hack (in all honesty) to simply make sure certain dict keys always get listed first in a sort.
    Assumes input is a list of keys (unique items) - will check.
    """
    if not keys or len(keys) < 1:
        return keys

    assert isinstance(keys, list), 'Expected a list for input to special_sort, got: '+str(keys)
    assert len(keys) == len(set(keys)), 'Input list cannot have repeats: '+str(keys)

    KEEP_FIRST = ['element_id', 'data_name'] # these are specific to validation constraint conditions
    retval = keys[:] # copy

    for item in KEEP_FIRST:
        if item in retval: # can only appear once so remove it and push to front
            retval.remove(item)
            retval = [item,]+retval

    return retval


def get_spaces(indent):
    if indent == None:
        return ''
    assert isinstance(indent, int), 'Expected int for indent, got: '+str(indent)
    return '    '*max(0,indent)


def to_str(thing, indent=None):
    """
    Stringfiy anything - dict or list.
    """
    if type(thing) == dict:
        return dict_to_str(thing, indent=indent)
    elif type(thing) == list:
        return list_to_str(thing, indent=indent)
    else:
        return '%s' % thing


def dict_to_str(adict, key_sorter=None, indent=None, breakup=False):
    """
    Turn Python dict into valid JS code; get around the unicode issue without requiring yaml, etc.
    To pretty-print, set indent to any integer.
    To simply break up long lines at each key, set breakup to True.
    Currently, key_sorter is NOT expected to be passed down recursively.
    """
    retval = '{'
    if indent != None:
        breakup = False # override
        retval += '\n'
    sep_space = '' if indent == None else ' '

    the_keys = sorted(adict, key=key_sorter)
    the_keys = special_sort(the_keys)
    for item in the_keys:
        val = adict[item]
        indent_arg = None if indent == None else indent+1
        if type(val) == dict:
            retval += get_spaces(indent)+"'"+item+"':"+sep_space+dict_to_str(val, indent=indent_arg)+','
            if indent != None: retval += '\n'
        elif type(val) == list:
            if len(val) == 0 or type(val[0]) in (str, int, float): indent_arg = None # be a more compact-pretty
            retval += get_spaces(indent)+"'"+item+"':"+sep_space+list_to_str(val, indent=indent_arg)+','
            if indent != None: retval += '\n'
        elif type(val) in (str,):
            retval += get_spaces(indent)+"'"+item+"':"+sep_space+"'"+val+"',"
            if indent != None: retval += '\n'
        else:
            retval += get_spaces(indent)+"'"+item+"':"+sep_space+str(val)+','
            if indent != None: retval += '\n'
        if breakup:
            retval += '\n'

    if retval.endswith('\n'): retval = retval[:-1] # rm poss. final newline
    if retval.endswith(',') : retval = retval[:-1] # rm final comma
    if indent != None: retval += '\n'
    if indent != None: indent -= 1 # tricky
    retval += get_spaces(indent)+'}'

    return retval.strip()


def list_to_str(alist, indent=None):
    """
    Turn Python list into valid JS code; get around the unicode issue without requiring yaml, etc.
    To pretty-print, set indent to any integer.
    """
    retval = "["
    if indent != None: retval += '\n'
    sep_space = '' if indent == None else ' '

    for item in alist:
        indent_arg = None if indent == None else indent+1
        if type(item) in (int, float):
            retval += get_spaces(indent)+str(item)+","
            if indent != None: retval += '\n'
        elif type(item) in (list, tuple):
            retval += get_spaces(indent)+list_to_str(item, indent=indent_arg)+","
            if indent != None: retval += '\n'
        elif type(item) == dict:
            if len(dict_to_str(item)) < 75: indent_arg = None # be a more compact-pretty
            retval += get_spaces(indent)+dict_to_str(item, indent=indent_arg)+","
            if indent != None: retval += '\n'
        else:
            retval += get_spaces(indent)+"'"+str(item)+"',"
            if indent != None: retval += '\n'

    if retval.endswith('\n'): retval = retval[:-1] # rm poss. final newline
    if retval.endswith(',') : retval = retval[:-1] # rm final comma
    if indent != None: retval += '\n'
    if indent != None: indent -= 1 # tricky
    retval += get_spaces(indent)+"]"

    # add some spacing simply to match previous format (can rm this code
    # in a separate stand-alone pr if we like), for code diffing
    if retval.startswith('[['): retval = '[ ' + retval[1:]
    if retval.endswith(']]'): retval = retval[:-1] + ' ]'

    return retval


def unquote_js_varnames(blob_string):
    """ Handle when something like pandeia.const.ERROR had to be held
    in Python as a string (cuz it's Python not JS in here), but should
    be output in JS as a var name not a string.
    """
    pattern = r"""["']pandeia\.const\.[A-Z][A-Z_]*["']"""
    retval = blob_string[:] # copy

    match = re.search(pattern, retval)
    while None != match:
        # got a single match, find its actual value and replace all occurances of it
        value = retval[match.start() : match.end()] # e.g. "pandeia.const.ERROR"
        better = value[1:-1]                        # e.g.  pandeia.const.ERROR
        retval = retval.replace(value, better)
        # iterate (find another, like say maybe "pandeia.const.WARNING")
        match = re.search(pattern, retval)

    return retval


def find_disp_str(disp_str_dict, itemname, secondary_key=None):
    """ Convenience - do a simple lookup in the dict but also cleanly handle
    when/if the disp_str_dict is absent, or doesn't contain itemname. If a secondary_key
    is provided and the display_string is actually a dict, then use it to lok up a value. """
    # if no possible display string, use key name itself, uppercased
    if disp_str_dict==None:
        return str(itemname).upper()

    if 'display_string' in disp_str_dict:
        val = disp_str_dict['display_string']
        if type(val) == dict:
            if secondary_key in val:
                return val[secondary_key];
            elif 'default' in val:
                return val['default']
            else:
                raise ValueError('display_string is mal-formed dict: '+str(val))
        else:
            return val

    elif itemname in disp_str_dict:
        return disp_str_dict[itemname]
    else:
        return str(itemname).upper()


def read_full_file(fname):
    """ convenience function """
    assert os.path.exists(fname), 'read_full_file expected, but did not find file named: '+fname
    f = open(fname, 'r')
    buf = f.read()
    f.close()
    return buf


def consolidate_bloated_expansion_dict(the_dict):
    """ There are cases where the expansion dict lists every option by aperture, but the choices
    are actually the same for each aperture, so simply consolidate that situation for clarity.
    """
    assert the_dict and 'pars' in the_dict and 'values' in the_dict, 'Malformed input to consolidate_bloated_expansion_dict()'
    # handle simple cases first
    if '*' in the_dict['values'] or len(the_dict['values']) == 1:
        # this dict is already consolidated enough, so do nothing
        return the_dict
    assert len(the_dict['values']) > 1, 'Programming error in consolidate_bloated_expansion_dict()'
    sample_aper = list(the_dict['values'])[0]
    # see if every choice set is the same as the choices for sample_aper
    choices_str = list_to_str(the_dict['values'][sample_aper])
    all_are_the_same = True
    for aper in the_dict['values']:
        this_choices_str = list_to_str(the_dict['values'][aper])
        if this_choices_str != choices_str:
            all_are_the_same = False
            break
    if not all_are_the_same:
        # ok, this was not bloat, we actually need to specify them all
        return the_dict
    else:
        # it is needlessly bloated. get rid of all keys but '*'
        the_only_choice_set = copy.deepcopy(the_dict['values'][sample_aper])
        the_dict['values'] = { '*': the_only_choice_set }
    return the_dict


def ironout_display_string_list(partly_nested_list):
    """ In case of display string list only is nested for some of its elements, 
    then it is better to iron them out and map the rest by doing camel case
    so that the whole list is nested.
    Example input: ['low', 'high', ['extremely_high', 'Extremely High']]
    Example output: [['low','Low'],['high','High'],['extremely_high','Extremely High']]
    """
    ironed_list = []
    for element in partly_nested_list:
        ironed_list.append([element, camel_case(element)] if type(element) is not list else element)
    return ironed_list


def generate_cfg_basics(pandeia_data_loc, telescope, all_output, verbose):
    """ Given the values found in json config files under
    pandeia_data, generate all needed javascript code lines for
    the client, which will later be dumped to a file.

    This currently LARGE function needs to be broken up more, for sure,
    but that can now be done anytime, now that "all_output" can
    be used/appended from function call to function call.

    Parameters
    ----------
    pandeia_data_loc: str
        Full path to pandeia_data tree to be used

    telescope: str
        Name of telescope/mission, e.g. 'jwst'

    all_output: dict
        The all_output dict is a IO param, returning info to caller.

    verbose: bool
        chatty versus not
    """

    # init
    instruments = sorted(SUPPORTED_MODES[telescope])
    inst_labels = { inst: inst.upper().replace('SPEC','Spec').replace('CAM','Cam') for inst in instruments }
    if verbose:
        print(f'telescope "{telescope}" has instruments: {instruments}')

    # order the output. this list contains either key-names or comment lines
    if 'order' not in all_output: all_output['order'] = []

    all_output['order'] += [
        ('supported_lines',      'What is supported:'),
        ('readpatt_lines',       'Readout pattern values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('subarray_lines',       'Subarray values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('enum_lines',           'Time value enumerated choices: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('sizes_lines',          'Physical sizes:'),
        ('constraint_lines',     'Constraints: each is a dict of dicts, top level key is aperture'),
        ('aperture_lines',       'Aperture values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('detector_lines',       'Detector info: mapping of aperture to detector values'),
        ('submode_lines',        'Sub-mode values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('dsprsr_lines',         'Disperser values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('filter_lines',         'Filter values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('dbl_filter_lines',     'Double filter values: each is a dict of { filter-to-pass-to-engine: filter-paired-with }'),
        ('cenwave_lines',        'Cenwave values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('slit_lines',           'Slit values: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('range_lines',          'Wavelength ranges (telescope_instr_mode_aperture): each line is a dict of: { filt/disp1: [wmin, wmax], filt/disp2: ... }'),
        ('shutlocs_lines',       'Shutter location choices: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('slitlet_lines',        'Slitlet choices: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('strategy_lines',       'Permitted strategies:'),
        ('stratname_lines',      'Strategy names:'),
        ('normalzn_lines',       'Normalization photsys: each is a list of lists: [ [value1,display_string1], [v2,ds2], ...]'),
        ('sed_lines',            'SED "groupings" and "families" (lists and list of dicts)'),
        ('extinct_lines',        'Extinction values'),
        ('glow_lines',           'Glow region values is by telescope_instrument_detector:  [ [v1,ds1], [v2,ds2], ...] ]'),
        ('exposure_conf_lines',  'Exposure calculation reference values'),
        ('bg_lines',             'Background standard levels: [ [value1,display_string1], [v2,ds2], ...]'),
    ]

    # to collect items - create empty lists
    for key_tup in all_output['order']:
        all_output[key_tup[0]] = []

    # meta info - can use simple conversion via str() here
    all_output['supported_lines'].append(f'CFG_VALS.SUPPORTED_{telescope}_instruments = {instruments};')
    all_output['supported_lines'].append(f'CFG_VALS.SUPPORTED_{telescope}_inst_labels = {dict_to_str(inst_labels)};')
    all_output['supported_lines'].append(f'CFG_VALS.SUPPORTED_{telescope}_modes = {dict_to_str(SUPPORTED_MODES[telescope])};')
    all_output['supported_lines'].append(f'CFG_VALS.SUPPORTED_{telescope}_spec_modenames = {SPEC_MODENAMES[telescope]};')
    all_output['supported_lines'].append(f'CFG_VALS.UNFINISHED_{telescope}_modes = {dict_to_str(UNFINISHED_MODES[telescope])};')
    if telescope == 'hst': # add jwst SUPPORTED_MODES
        all_output['supported_lines'].append(f'CFG_VALS.SUPPORTED_jwst_instruments = {sorted(SUPPORTED_MODES["jwst"])};')
        all_output['supported_lines'].append(f'CFG_VALS.SUPPORTED_jwst_modes = {dict_to_str(SUPPORTED_MODES["jwst"])};')

    # for hst background choices
    if telescope == 'hst':
        all_output['bg_lines'].append(f'CFG_VALS.BG_CHOICES_hst_airglow = {to_str(ironout_display_string_list(HST_BG_AIRGLOW_CHOICES))};')
        all_output['bg_lines'].append(f'CFG_VALS.BG_CHOICES_hst_earthshine = {to_str(ironout_display_string_list(HST_BG_EARTHSHINE_CHOICES))};')
        all_output['bg_lines'].append(f'CFG_VALS.BG_CHOICES_hst_zodiacallight = {to_str([[s, camel_case(s)] for s in HST_BG_ZODIACALLIGHT_CHOICES])};')

    # loop through all supported instruments and modes
    for instr in instruments:

        # open the instrument config file
        fname = f'{pandeia_data_loc}/{telescope}/{instr}/config.json'
        cfg_tree = json.loads(read_full_file(fname))
        assert type(cfg_tree) == dict, 'Unexpected json layout in: '+fname
        strip_meta(cfg_tree)
        if verbose:
            print(f'reading {instr} data from: {fname}')

        # mode labels for this instrument (allow this to throw KeyError if data is partial)
        try:
            mode_labels = { mode:cfg_tree['mode_config'][mode]['display_string'] for mode in SUPPORTED_MODES[telescope][instr] }
        except Exception as e:
            print(instr, mode_labels)
            raise e

        all_output['supported_lines'].append(f'CFG_VALS.SUPPORTED_{telescope}_'+instr+'_mode_labels = '+dict_to_str(mode_labels)+';')

        # Grab values required for front-end exposure time calculations
        line = f'CFG_VALS.SUBARRAYS_CONF_{telescope}_{instr} = {dict_to_str(cfg_tree["subarray_config"])};'
        all_output['exposure_conf_lines'].append(line)

        # Create the mapping of aperture-to-detector for this instrument (assume can exist for all missions)
        aper2detr = {aper: cfg_tree['aperture_config'][aper]['detector'] for aper in cfg_tree['aperture_config']}
        line = f'CFG_VALS.APER2DET_{telescope}_{instr} = {dict_to_str(aper2detr)};'
        all_output['detector_lines'].append(line)

        # grab detector specifics for each detector
        for detr in cfg_tree['detector_config']:
            ramp_cfg = {key:cfg_tree['detector_config'][detr][key] for key in cfg_tree['detector_config'][detr] if key in ['npostrej','nprerej','nreset']}
            # for now for JWST key this by aperture since the client knows aper and since there
            # is currently a 1-1 corresponence between detr and aper for JWST, and since HST has no ramp config
            # (may need to refactor this tho someday)
            line = f'CFG_VALS.RAMP_CONF_{telescope}_{instr}_{detr} = {dict_to_str(ramp_cfg)};'
            all_output['exposure_conf_lines'].append(line)

            # glow values (assumes that all have a 'none' choice)
            if 'dark_glow_region_none' in cfg_tree['detector_config'][detr]:
                # this is awkward but we need to find the possible values by looking at keys
                glow_choices = [k[17:] for k in cfg_tree['detector_config'][detr] if k.startswith('dark_glow_region_')]
                ordered_glow_choices = sorted(glow_choices, key=lambda x: "nlmh".find(x.lower()[0]))
                line = f'CFG_VALS.GLOW_VALS_{telescope}_{instr}_{detr} = {to_str([[s, camel_case(s)] for s in ordered_glow_choices])};'
                all_output['glow_lines'].append(line)

        line = f'CFG_VALS.READMODE_CONF_{telescope}_{instr} = {dict_to_str(cfg_tree["readout_pattern_config"])};'
        all_output['exposure_conf_lines'].append(line)

        # aperture/shutter/slit/etc sizes
        # cfg_tree['aperture_config']['shutter']
        if 'aperture_config' in cfg_tree:
            aper_tree = cfg_tree['aperture_config']

        if 'slit_config' in cfg_tree:
            slit_tree = cfg_tree['slit_config']
            # Get info on any instrument that has a shutter
            if 'shutter' in slit_tree:
                height = slit_tree['shutter']['xdisp']
                width  = slit_tree['shutter']['disp']
                all_output['sizes_lines'].append(f'CFG_VALS.SIZES_{telescope}_{instr}_shutter_height = {height};')
                all_output['sizes_lines'].append(f'CFG_VALS.SIZES_{telescope}_{instr}_shutter_width = {width};')

                # Read shutter cfg file
                fname2 = f'{pandeia_data_loc}/{telescope}/{instr}/shutters.json'
                shutter_tree = json.loads(read_full_file(fname2))
                strip_meta(shutter_tree)
                shutlocs = [ [k,shutter_tree[k]['display_string']] for k in sorted(shutter_tree) ]
                all_output['shutlocs_lines'].append(f'CFG_VALS.SHUTTER_LOCS_{telescope}_{instr} = {list_to_str(shutlocs)};')

        # mode-specific items
        for mode in sorted(SUPPORTED_MODES[telescope][instr]):
            # basics
            instr_mode = instr+'_'+mode
            if verbose: print(instr_mode)

            # mode config sub-dict
            mode_cfg = {}
            if 'mode_config' in cfg_tree:
                mode_cfg = cfg_tree['mode_config'][mode]

            if telescope != 'hst':
                # readout_patterns
                where2look = mode_cfg if 'readout_patterns' in mode_cfg else cfg_tree
                rdmd_info = [ [rdmd,find_disp_str(cfg_tree['readout_pattern_config'][rdmd],rdmd,mode)] for rdmd in where2look['readout_patterns'] ]
                line = f'CFG_VALS.READMODES_{telescope}_{instr}_{mode} = {list_to_str(rdmd_info)};'
                all_output['readpatt_lines'].append(line)

                # subarray
                where2look = mode_cfg if 'subarrays' in mode_cfg else cfg_tree
                # Use "default" only here for now since _this_ CFG_VALS entry is only used in the UI for having the
                # full per-instr_mode list of subarrays (doesn't depend on readout_pattern).  Note however that SUBARRAYS_CONF_* has
                # the full tree of readout_pattern-dependent information on subarrays
                sbry_info = [ [sbry,find_disp_str(cfg_tree['subarray_config']['default'][sbry],sbry,mode)] for sbry in where2look['subarrays'] ]
                line = f'CFG_VALS.SUBARRAYS_{telescope}_{instr}_{mode} = {list_to_str(sbry_info)};'
                all_output['subarray_lines'].append(line)

            # enumerated time items (optional)
            for itemname, capname in [('enum_nexps','NEXPS'),('enum_ngroups','NGROUPS'),('enum_nints','NINTS')]:
                where2look = mode_cfg if itemname in mode_cfg else cfg_tree
                if itemname in where2look:
                    if 'key_format' in where2look[itemname]: # it is broken out by some other parameter
                        # the dict of enum_item_info as is, where2look[itemname], is almost complete but
                        # we need to delve into each value (array) to assign display_string's
                        enum_item_info = { k:[[v, str(v)] for v in where2look[itemname][k]] for k in where2look[itemname] if k != 'key_format'}
                        enum_item_info['key_format'] = where2look[itemname]['key_format'] # handle this separately
                        line = f'CFG_VALS.{capname}_{telescope}_{instr}_{mode} = {dict_to_str(enum_item_info)};'
                    else: # simple list of values - just assign display_string's now
                        enum_item_info = [ [item, str(item)] for item in sorted(where2look[itemname], key=lambda n: float(n)) ]
                        line = f'CFG_VALS.{capname}_{telescope}_{instr}_{mode} = {list_to_str(enum_item_info)};'
                    all_output['enum_lines'].append(line)

            # constraints (e.g. miri coronagraphy)
            if instr_mode in CONSTRAINED_MODES[telescope] and 'config_constraints' in cfg_tree:
                # History: so, up until Nov 2021, the only primary key here had ever been 'apertures', but now
                # nirspec (fs and ta) and stis (spec modes) allow both 'apertures' and 'dispersers' to be the primary key.
                # What to do?  To keep the client from getting too crazy before we need it to, we will still allow
                # only one primary constraint key per mode, but we will decide in this tool what it is.  For now...
                primary_c_key = 'dispersers' if instr_mode in CONSTRAINED_MODES_VIA_DISERSERS[telescope] else 'apertures'
                constraint_tree = cfg_tree['config_constraints'][primary_c_key]

                pck_vals = mode_cfg[primary_c_key] # eg either all the apertures or all the dispersers

                # NOTE - Sanity Check all constraints !
                # We actually force that all pck_vals values MUST exist in the constraint_tree for a given
                # thing (e.g. filters), or no pck_vals at all for that thing, bc otheriwse the JS logic in
                # the UI (selected_aperture() func) becomes too unwieldy
                for constrained_thing in ['cenwaves','filters','readout_patterns','slits','subarrays']: # add dispersers later? (needs work)
                    constrained_thing_keys = [k for k in constraint_tree if constrained_thing in constraint_tree[k]]
                    if constrained_thing_keys == [] or \
                       (instr_mode, constrained_thing) in IGNORE_CONSTRAINT_SANITY_CHECKS or \
                       set(pck_vals).issubset(constrained_thing_keys):
#                       print('Constraints seem sane/correct for: '+instr_mode+'/'+constrained_thing)
                        pass
                    else:
                        # the constraints for this (instr_mode, thing) need to be checked and may have a self-consistency issue ...
                        if len(constrained_thing_keys) > 0 and set(constrained_thing_keys).issubset(pck_vals):
                            # ahh - simple case - we are missing choice-lists for some apers e.g., and
                            # so we can simply/safely populate with the mode_config set of choices
                            keys_missing_constraints = tuple(set(pck_vals) - set(constrained_thing_keys))
                            all_choices_for_thing = copy.deepcopy(mode_cfg[constrained_thing])
                            if verbose:
                                print('\nWARNING: handling sparsely-populated constraints: '+instr_mode+'/'+constrained_thing+' ...')
#                           print('we will now allow each of these primary constraint key values: '+str(keys_missing_constraints))
#                           print('to each use all of these "'+constrained_thing+'" choices: '+str(all_choices_for_thing))
                            # Here we are actually adding to the constraint_tree read from the config file (but

                            # of course not editing the config file), to fill in missing choices for sparsely-populated
                            # constraint information.  If an instr_mode is NOT constrained, it gets all valid choices
                            # of "thing" (e.g. subarray) from mode_config.
                            for key_missing_constraints in keys_missing_constraints:
                                assert key_missing_constraints not in constraint_tree, '??? cant happen, see above ???'
                                constraint_tree[key_missing_constraints] = {constrained_thing: {'default': all_choices_for_thing}}
#                               print('Added '+constrained_thing+' constraint for: '+key_missing_constraints+ \
#                                     ':\n\t'+str(constraint_tree[key_missing_constraints]))
                        else:
                            raise ValueError('Problem with '+instr_mode+' constraints on: '+constrained_thing+ \
                                             '\nFull PCK list:  '+str(sorted(pck_vals))+ \
                                             '\nconstraint keys: '+str(sorted(constrained_thing_keys))+ \
                                             '\n\nwhole constraint dict: '+str(constraint_tree) )

                # "matches" is the dict we primarily work with here, but it currently shares references
                # with cfg_tree; since we are going to modify matches in place, but don't want to
                # modify cfg_tree or mode_cfg, we make a deep copy of it and use that instead.
                matches = copy.deepcopy({k:constraint_tree[k] for k in pck_vals})
                # add the display_string info - modify dict in place
#               for k in list(matches): # modifying dict so copy keys
#                   matches[k]['display_string'] = find_disp_str(aper_tree[k], k, mode)
                # also resolve some of the values (where we usu. expect lists) to see if they set a level
                # lower so they can vary by ins.mode (if so, collapse to just lists using ins.mode);
                # modify dict in place
                for k in list(matches): # modifying dict so copy keys
                    this_pckeys_constraints = matches[k]
                    for constraint_name in list(this_pckeys_constraints): # modifying dict so copy keys
                        constraint_val = this_pckeys_constraints[constraint_name]
                        # here is where we resolve the choice and reduce the tree by a level
                        if type(constraint_val) == dict:
                            if mode in constraint_val:
                                new_val = constraint_val[mode]
                            elif 'default' in constraint_val:
                                new_val = constraint_val['default']
                            else:
                                raise ValueError('constraint sub-dict is a mal-formed dict: '+str(constraint_val))
                            # Check for the USE-THEM-ALL flag by seeing if they set this val to null/None
                            if new_val == None:
                                # they said "null", which means "no constraint", so use the list (e.g. of filters)
                                # which applies to the entire mode (independently of primary constraint key)
                                if constraint_name in mode_cfg:
                                    new_val = copy.deepcopy(mode_cfg[constraint_name])
                                else:
                                    # at this point (e.g. say for filters) we would need to check if the global
                                    # instrument-level (file-level) list is set and use that; add that code here
                                    raise ValueError('constraint "'+constraint_name+ \
                                          '" needs a fallback set of choices but none were found in mode_config')
                            # dict update
                            this_pckeys_constraints[constraint_name] = new_val
                    # update the matches dict
                    matches[k] = this_pckeys_constraints

                # any last-minute hacks
                matches = hacks_for_ts_dblfilts(mode, matches)

                # Make any translations for the readmode-to-readout_pattern name change.
                # The data and engine are now using 'readout_pattern', but the client is still
                # using 'readmode' - and we cannot fully convert the client internal names
                # and variables until we are will to do some very extensive manual testing.
                junk = {k: { 'readmodes' if k2=='readout_patterns' else k2 : matches[k][k2] for k2 in matches[k]} for k in matches}
                matches = junk

                # done
                line = f'CFG_VALS.CONSTRAINTS_{telescope}_{instr_mode} = {dict_to_str(matches)};'
                all_output['constraint_lines'].append(line)
                all_output['CACHED'][f'CONSTRAINTS_{telescope}_{instr_mode}'] = matches # stash away raw unstringified data for later use

                singular_primary_c_key = primary_c_key[0:-1] if primary_c_key.endswith('s') else primary_c_key
                line = f'CFG_VALS.CONSTRAINTS_KEY_{telescope}_{instr_mode} = "{singular_primary_c_key}";'
                all_output['constraint_lines'].append(line)

            # apertures - see if there are specific apertures allowed for this mode
            valid_apertures = []
            if 'apertures' in mode_cfg:
                valid_apertures = mode_cfg['apertures']
                aperinfo = [ [a, find_disp_str(aper_tree[a], a, mode)] for a in valid_apertures ]
                # any last-minute hacks
                aperinfo = hacks_for_ts_apertures(mode, aperinfo[:])
                # write it out and cache it
                line = f'CFG_VALS.APERTURES_{telescope}_{instr_mode} = {to_str(aperinfo)};'
                all_output['aperture_lines'].append(line)
                all_output['CACHED'][f'APERTURES_{telescope}_{instr_mode}'] = aperinfo # stash away raw unstringified data for later use

            # submodes - see if there are specific sub-modes allowed for this mode (also sometimes called "flavor")
            valid_submodes = []
            if 'submodes' in mode_cfg:
                raise ValueError('THIS IS NOT YET IMPLEMENTED')
            else:
                if instr_mode == 'nircam_sw_ts':
                    submodeinfo = hacks_for_sw_ts_submodes()
                    # write it out and cache it
                    line = f'CFG_VALS.SUBMODES_{telescope}_{instr_mode} = {to_str(submodeinfo)};'
                    all_output['submode_lines'].append(line)
                    all_output['CACHED'][f'SUBMODES_{telescope}_{instr_mode}'] = submodeinfo # stash away raw unstringified data for later use
                    # also make a submode mapping - to be an easy inverse lookup for the client
                    submodemap = {}
                    for submode in all_output['CACHED'][f'APERTURES_{telescope}_{instr_mode}']: # cuz this is a dict & the key is submode
                        apers = [a[0] for a in all_output['CACHED'][f'APERTURES_{telescope}_{instr_mode}'][submode]]
                        submodemap.update({k:submode for k in apers})
                    # write it out and cache it
                    line = f'CFG_VALS.SUBMODE_MAP_{telescope}_{instr_mode} = {to_str(submodemap)};'
                    all_output['submode_lines'].append(line)
                    all_output['CACHED'][f'SUBMODE_MAP_{telescope}_{instr_mode}'] = submodemap

            # wavelength ranges (filter, disperser, etc)
            if mode in SPEC_MODENAMES[telescope] and 'range' in cfg_tree and len(valid_apertures)>0:
                for va in valid_apertures:
                    # For most inst/modes we use the simpler tree: cfg_tree['range'][va], but for
                    # disp-filt using modes we will use:
                    #    aperture_config.<aper>.gap.<default|sub_key>.<grism>.*|<filter>
                    #
                    if instr_mode in DISPFILT_MODES[telescope] and instr_mode != "nirspec_msa": # shutters are more work, #4645
                        assert va in cfg_tree['aperture_config'] and cfg_tree['aperture_config'][va]['slit'] in cfg_tree['slit_config'] and 'gap' in cfg_tree['slit_config'][cfg_tree['aperture_config'][va]['slit']], 'malformed aper cfg for '+mode
                        # Get range info from REFWAVE_DICT, but convert it to the format we want
                        full_aper_wave_tree = {}
                        rwdimv = REFWAVE_DICT[instr][mode][va]
                        # Then convert it to the format we need
                        for akey in rwdimv:
                            if isinstance(rwdimv[akey], dict):
                                if 'wmin' in rwdimv[akey]:
                                    full_aper_wave_tree[akey] = [ rwdimv[akey]['wmin'], rwdimv[akey]['wmax'] ]
                                else: # it is a dict of subdicts
                                    full_aper_wave_tree[akey] = {}
                                    for akey2 in rwdimv[akey]:
                                        full_aper_wave_tree[akey][akey2] = [ rwdimv[akey][akey2]['wmin'], rwdimv[akey][akey2]['wmax'] ]
                            else:
                                full_aper_wave_tree[akey] = rwdimv[akey]
                        line = f'CFG_VALS.WAVE_RANGES_{telescope}_{instr_mode}_{va} = {dict_to_str(full_aper_wave_tree)};'
                    else:
                        ranges = cfg_tree['range'][va] # exists in cfg stuff as a dict
                        # get the range keys for this aper, do not include the aper-wide fallback values
                        range_keys = [k for k in ranges if k not in ('wmin','wmax')]

                        # if there are keys in the range_keys list, create a dict; if not, there is a single name-less key in the dict
                        if len(range_keys):
                            # random spot-check
                            assert ranges[range_keys[0]]['wmax'] > ranges[range_keys[0]]['wmin'], 'Invalid wmin,wmax: '+str(ranges[range_keys[0]])
                            # get arrays of ranges, in keyed dicts (key if usu. filter name but varies by mode)
                            range_arrays = {}
                            for rkey in range_keys:
                                subkeys = [k for k in ranges[rkey] if k not in ('wmin','wmax')]
                                if subkeys and type(ranges[rkey][subkeys[0]]) is dict:
                                    # This is a setup which is one level deeper than ususal
                                    for subkey in subkeys:
                                        range_arrays[rkey+','+subkey] = [ranges[rkey][subkey]['wmin'], ranges[rkey][subkey]['wmax']]
                                else:
                                    # This is the normal setup, so grab the wmin and wmax.  But be aware
                                    # that MULTIORDER_MODES use a double key in the client, and handle it.
                                    rkeyfix = rkey if instr_mode not in MULTIORDER_MODES[telescope] else rkey+',*'
                                    range_arrays[rkeyfix] = [ranges[rkey]['wmin'], ranges[rkey]['wmax']]
                        else:
                            # in this case ranges had ONLY the fallback values
                            assert ','.join(sorted(ranges)) == 'wmax,wmin', 'Unexpected ranges object: '+str(ranges)
                            range_arrays = { '': [ranges['wmin'],ranges['wmax']] }

                        line = f'CFG_VALS.WAVE_RANGES_{telescope}_{instr_mode}_{va} = {dict_to_str(range_arrays)};'
                    all_output['range_lines'].append(line)

            # filters: is a listlist to preserve key order; is why we can't make it a dict/Obj { filtval: display_str, ... }
            filtinfo = []
            dispinfo = []
            cenwinfo = []
            if instr_mode in DISPERSER_MODES[telescope]:
                where2look = mode_cfg if 'dispersers' in mode_cfg else cfg_tree
                dispinfo = [ [g, find_disp_str(cfg_tree['disperser_config'][g], g)] for g in where2look['dispersers'] ]
                line = f'CFG_VALS.DISPERSERS_{telescope}_{instr_mode} = {list_to_str(dispinfo)};'
                all_output['dsprsr_lines'].append(line)
                all_output['CACHED'][f'DISPERSERS_{telescope}_{instr_mode}'] = dispinfo # stash away raw unstringified data for later use
            if instr_mode in DISPFILT_MODES[telescope]:
                if 'config_constraints' in cfg_tree:
                    dispcfg = cfg_tree['config_constraints']['dispersers']
                    for disp in [d for d in sorted(dispcfg) if 'filters' in dispcfg[d]]:
                        for filt in dispcfg[disp]['filters']:
                            disp_string = disp.upper() if 'disperser_config' not in cfg_tree else cfg_tree['disperser_config'][disp]['display_string']
                            filt_string = filt.upper() if    'filter_config' not in cfg_tree else cfg_tree['filter_config'][filt]['display_string']
                            filtinfo.append( [disp+','+filt, disp_string+'/'+filt_string] )
            else:
                filtinfo = [ [f, find_disp_str(cfg_tree['filter_config'][f], f, mode)] for f in mode_cfg['filters'] ]

            line = f'CFG_VALS.FILTERS_{telescope}_{instr_mode} = {list_to_str(filtinfo)};'
            all_output['filter_lines'].append(line)
            all_output['CACHED'][f'FILTERS_{telescope}_{instr_mode}'] = filtinfo # stash away raw unstringified data for later use

            # double filters
            if 'double_filters' in cfg_tree and 'filters' in mode_cfg:
                dbl_filts = [f for f in cfg_tree['double_filters'] if f in mode_cfg['filters']]
                if dbl_filts:
                    dblfiltinfo = { df: cfg_tree['double_filters'][df] for df in dbl_filts }
                    line = f'CFG_VALS.DOUBLE_FILTERS_{telescope}_{instr_mode} = {dict_to_str(dblfiltinfo)};'
                    all_output['dbl_filter_lines'].append(line)
                    all_output['CACHED'][f'DOUBLE_FILTERS_{telescope}_{instr_mode}'] = filtinfo # stash away raw unstringified data for later use

            # cenwaves
            if 'cenwaves' in mode_cfg:
#               cenwinfo = [ [x, find_disp_str(cfg_tree['cenwave_config'][x], x, mode)] for x in mode_cfg['cenwaves'] ]
                # !!! fix when the data is better for STIS
                cenwinfo = [ [x, find_disp_str(cfg_tree['cenwave_config'].get(x, x.upper()), x, mode)] for x in mode_cfg['cenwaves'] ]
                line = f'CFG_VALS.CENWAVES_{telescope}_{instr_mode} = {list_to_str(cenwinfo)};'
                all_output['cenwave_lines'].append(line)
                all_output['CACHED'][f'CENWAVES_{telescope}_{instr_mode}'] = cenwinfo

            # slits
            if 'slits' in mode_cfg:
                slitinfo = [ [x, find_disp_str(cfg_tree['slit_config'][x], x, mode)] for x in mode_cfg['slits'] ]
                line = f'CFG_VALS.SLITS_{telescope}_{instr_mode} = {list_to_str(slitinfo)};'
                all_output['slit_lines'].append(line)
                all_output['CACHED'][f'SLITS_{telescope}_{instr_mode}'] = slitinfo

            # strategies
            stratcfg = cfg_tree['strategy_config'][mode]
            line = 'CFG_VALS.STRATEGIES_'+instr_mode+' = '+ \
                   list_to_str([s for s in stratcfg['permitted_methods'] if s not in NON_UI_STRATEGIES])+';'
            all_output['strategy_lines'].append(line)

            # slitlets
            if 'slitlets' in mode_cfg:
                slitlet_info = [ [slitlet,find_disp_str(cfg_tree['slitlet_config'][slitlet],slitlet,mode)] for slitlet in mode_cfg['slitlets'] ]
                line = f'CFG_VALS.SLITLETS_{telescope}_{instr_mode} = {list_to_str(slitlet_info)};'
                all_output['slitlet_lines'].append(line)

    # RELATED HST INFO: loop through all supported modes (types->hst->bandpasses)
    # NOTE - the HST info is currently kindof hacked - needs to be revisited !
    if telescope != 'hst':
        RELATED_HST_INSTRUMENTS = hack_hst_cfg('instruments')
        all_output['supported_lines'].append("CFG_VALS.RELATED_hst_instruments = "+list_to_str(RELATED_HST_INSTRUMENTS)+';')
        RELATED_HST_MODES = hack_hst_cfg('modes')
        all_output['supported_lines'].append("CFG_VALS.RELATED_hst_modes = "+dict_to_str(RELATED_HST_MODES)+';')

        fname = pandeia_data_loc+'/normalization/config.json'
        cfg_tree = json.loads(read_full_file(fname))
        assert type(cfg_tree) == dict, 'Unexpected json layout in: '+fname
        assert 'types' in cfg_tree
        assert 'hst' in cfg_tree['types']
        assert 'bandpasses' in cfg_tree['types']['hst']
        bpdict = cfg_tree['types']['hst']['bandpasses']
        bandpasses = [b for b in sorted(bpdict) if 'display_string' in bpdict[b]]

        bp_collection_dict = {}
        for bp in bandpasses:
            bpinst,   bpmode,   bpfilt    = bp.split(',')
            bpinst_ds,bpmode_ds,bpfilt_ds = bpdict[bp]['display_string'].split('/')

            assert bpinst in RELATED_HST_INSTRUMENTS
            assert bpmode in RELATED_HST_MODES[bpinst]
            key = 'FILTERS_hst_'+bpinst+'_'+bpmode
            if key in sorted(bp_collection_dict):
                bp_collection_dict[key].append([bpfilt,bpfilt_ds])
            else:
                bp_collection_dict[key] = [ [bpfilt,bpfilt_ds], ]

        for key in sorted(bp_collection_dict):
            line = "CFG_VALS."+key+" = "+list_to_str(bp_collection_dict[key])+';'
            all_output['filter_lines'].append(line)
            all_output['CACHED'][key] = bp_collection_dict[key] # stash away raw unstringified data for later use

    # Normalization values used by the UI
    fname = pandeia_data_loc+'/normalization/config.json'
    cfg_tree = json.loads(read_full_file(fname))
    assert type(cfg_tree) == dict, 'Unexpected json layout in: '+fname
    systems = sorted(cfg_tree['types']['photsys']['bandpasses'])
    systemsinfo = [ [s, find_disp_str(cfg_tree['types']['photsys']['bandpasses'][s],s) ] for s in systems]
    all_output['normalzn_lines'].append("CFG_VALS.NORM_photsys = "+list_to_str(systemsinfo)+';')
    alpha_ordering_fam = [ "msx" ]
    for sstm in systems:
        filts = [str(f) for f in sorted(cfg_tree['types']['photsys']['bandpasses'][sstm].keys()) if f != 'display_string']
        # added str() call in list comprehension above to rm unicode-ness for py2
        # if not in alphabetical ordering family, sort by photometric system passbands 
        if sstm not in alpha_ordering_fam:
            filts = sorted(filts, key=lambda x: "ubvgrizyjhklmnq".find(x.lower()[0]))
        filtinfo = [ [f,cfg_tree['types']['photsys']['bandpasses'][sstm][f]['display_string']] for f in filts]
        all_output['normalzn_lines'].append("CFG_VALS.NORM_photsys_"+sstm+" = "+list_to_str(filtinfo)+';')
    unitsinfo = [ [u, u] for u in cfg_tree['defaults']['fluxunits'] ]
    all_output['normalzn_lines'].append("CFG_VALS.NORM_fluxunits = "+list_to_str(unitsinfo)+';')
    unitsinfo = [ [u, u] for u in cfg_tree['types']['at_lambda']['defaults']['fluxunits'] ]
    all_output['normalzn_lines'].append("CFG_VALS.NORM_at_lambda_fluxunits = "+list_to_str(unitsinfo)+';')
    unitsinfo = [ [u, u] for u in cfg_tree['defaults']['waveunits'] ]
    all_output['normalzn_lines'].append("CFG_VALS.NORM_waveunits = "+list_to_str(unitsinfo)+';')

    # SED values used by the UI
    # !NOTE same logic in pandeia_test/test_utilities/client/regen_lib/vary_all_sources.py:get_all_sed_cases()
    fname = pandeia_data_loc+'/sed/config.json'
    cfg_tree = json.loads(read_full_file(fname))
    assert type(cfg_tree) == dict, 'Unexpected json layout in: '+fname
    groupings = [sg['name'] for sg in cfg_tree['sed_groupings'] if 'display_string' in sg]
    groupingsinfo = [ [groupings[i], find_disp_str(cfg_tree['sed_groupings'][i],groupings[i]) ] for i in range(len(groupings))]
    all_output['sed_lines'].append("CFG_VALS.SED_groupings = "+list_to_str(groupingsinfo)+';')

    bkwd_map = {}
    for grp_idx, grp in enumerate(groupings):
        # create a single dict for this grouping
        sedtypes = cfg_tree['sed_groupings'][grp_idx]['sed_types']  # e.g. for grp=analytics, family/sedtypes=["flat", "powerlaw", "blackbody"]
        grouping_dict = {}
        grouping_dict['MENU'] = [[fam, cfg_tree['sed_types'][fam]['display_string']] for fam in sedtypes]
        # now add a dict entry for every family in this grouping
        for family in sedtypes:
            special_ordering_fam = [ "phoenix" , "k93models" ]
            bkwd_map[family] = grp
            if 'spectra' in cfg_tree['sed_types'][family]:
                # open and read this spectra config file
                fname = pandeia_data_loc+'/sed/'+cfg_tree['sed_types'][family]['spectra']['config']
                spec_cfg_tree = json.loads(read_full_file(fname))
                assert type(spec_cfg_tree) == dict, 'Unexpected json layout in: '+fname
                ordered_keys = spec_cfg_tree.keys() # do not change orig json file sort order (json load preserves file order with py3.6? json)
                has_subfams = 'subfam' in spec_cfg_tree[(list(ordered_keys))[0]]
                subfams = {}
                if has_subfams:
                    for speckey in ordered_keys:
                        subfam = spec_cfg_tree[speckey]['subfam']
                        if subfam not in subfams:
                            # as of py3.7, dict keys will preserve insertion order (we rely on this for the subfams dict)
                            subfams[subfam] = []
                        subfams[subfam].append(speckey)

                if family in special_ordering_fam:
                    # 3 steps to get a deterministic sort the way we want it:
                    # 1) first order by alpha (and rm unicode-ness for py2)
                    keys_by_abc = sorted([str(k) for k in spec_cfg_tree.keys()], key=lambda k: str(spec_cfg_tree[k]))
                    # 2) then order by descending "teff"
                    # (also involve the log_g value as a tiny effect to break ties (make deterministic) when two items have same teff)
                    keys_by_teff = sorted(keys_by_abc, key=lambda k: spec_cfg_tree[k]['teff']+((100.-spec_cfg_tree[k]['log_g'])/1000.), reverse=True)
                    # 3) now sort first by OBA... w/ teff as secondary (and alpha as tertiary)
                    #      ... mnemonic is "Oh Be A Fine Girl, Kiss Me" (google it, I dare ya)
                    ordered_keys = sorted(keys_by_teff, key=lambda x: "obafgkmlty".find(x.lower()[0]))

                grouping_dict[family] = {}
                if has_subfams:
                    for subfam in subfams:
                        sf_specsinfo = [ [s, spec_cfg_tree[s]['display_string'] ] for s in subfams[subfam] if 'display_string' in spec_cfg_tree[s]]
                        grouping_dict[family][subfam] = sf_specsinfo
                    grouping_dict[family]['MENU'] = [ [s,s] for s in subfams ] # currently no special display_string vals for subfams
                else:
                    specsinfo = [ [s, spec_cfg_tree[s]['display_string'] ] for s in ordered_keys if 'display_string' in spec_cfg_tree[s]]
                    grouping_dict[family][""] = specsinfo

        all_output['sed_lines'].append("CFG_VALS.SED_families_"+grp+" = "+dict_to_str(grouping_dict, breakup=1)+';')

    all_output['sed_lines'].append("CFG_VALS.SED_family_to_grouping = "+dict_to_str(bkwd_map)+';') # allows us to have cleaner client logic

    # Strategy official names
    stnms = {}
    stltr = {}
    for stratfname in glob.glob(pandeia_data_loc+'/strategy/*.json'):
        strat_short = os.path.splitext(os.path.basename(stratfname))[0]
        strat_tree = json.loads(read_full_file(stratfname))
        strat_name = strat_tree['display_string']
        stnms[strat_short] = strat_name
        if strat_short not in NON_UI_STRATEGIES:
            stltr[strat_short] = STRATEGY_LETTERS[strat_short]
    all_output['stratname_lines'].append("CFG_VALS.STRATEGY_NAMES = "+dict_to_str(stnms)+';')
    all_output['stratname_lines'].append("CFG_VALS.STRATEGY_LETTERS = "+dict_to_str(stltr)+';')

    # Extinction information
    fname = pandeia_data_loc + '/extinction/config.json'
    cfg_tree = json.loads(read_full_file(fname))

    law_families = list(zip(sorted(cfg_tree['law_families']),
                       [cfg_tree['law_families'][x]['display_string']
                        for x in sorted(cfg_tree['law_families'])]))

    all_output['extinct_lines'].append(f"CFG_VALS.EXTINCT_law_family = {list_to_str(law_families)};")

    law_types = list(zip(sorted(cfg_tree['laws']),
                    [cfg_tree['laws'][x]['display_string']
                     for x in sorted(cfg_tree['laws'])]))

    for fam_id, fam_label in law_families:
        law_list = [x for x in law_types if x[0] in cfg_tree['law_families'][
            fam_id]['law_types']]

        all_output['extinct_lines'].append(f"CFG_VALS.EXTINCT_law_family_{fam_id} = {list_to_str(law_list)};")

    # sort by photometric system. e.g. "vjhk"
    sorted_bandpasses = sorted(cfg_tree['bandpasses'], key=lambda x: "ubvgrizyjhklmnq".find(x.lower()[0]))
    bands = list(zip(sorted_bandpasses,
                [cfg_tree['bandpasses'][x]['display_string'] for x in sorted_bandpasses]))

    all_output['extinct_lines'].append(f"CFG_VALS.EXTINCT_bandpasses = {list_to_str(bands)};")

    # done
    if verbose:
        print(f'Finished generate_cfg_basics() for telescope: {telescope}')


def get_possible_values_for(all_output, telescope, instr, mode, aper, parname):
    """ Return a list of the possible values for the parname given. Takes constraints into account. """
    # handle case where this was called for a non-existant par (e.g. the 3rd par when only all-comb'ing 2 pars)
    if parname == None:
       return [None]

    # sanity checks
    assert parname in ('aperture','disperser','filter'), 'get_possible_values_for not yet handling par named: '+parname

    instr_mode = instr+'_'+mode

    # now see if there were constraints on these values
    constraints_key = f'CONSTRAINTS_{telescope}_{instr_mode}'
    if aper != '*' and constraints_key in all_output['CACHED'] and aper in all_output['CACHED'][constraints_key]:
        # there are constraints for this instr_mode, so see if they affect the par of interest here.
        name_of_group = parname+'s' # for a key like 'filter', the constraint dict has the key 'filters'
        if name_of_group in all_output['CACHED'][constraints_key][aper]:
            # OK, so there _ARE_ constraints we need to worry about.  Grab all poss. values, for the given aper
            return all_output['CACHED'][constraints_key][aper][name_of_group]

    # if we get here there are no constraints to use, so collect/send all possible values for this par
    cache_key = f'{parname.upper()}S_{telescope}_{instr_mode}'
    assert cache_key in all_output['CACHED'], 'Key not found in running cache?!: %s' % cache_key
    if type(all_output['CACHED'][cache_key]) == dict:
        # cached info is a dict of lists
        vals = []
        for key in all_output['CACHED'][cache_key]:
            vals = vals + [item[0] for item in all_output['CACHED'][cache_key][key]]
        return vals
    else:
        # cahced info is a simple list of lists
        return [item[0] for item in all_output['CACHED'][cache_key]]


def generate_expansions(pandeia_data_loc, telescope, all_output):
    """ Given the values found in json config files under
    pandeia_data, generate the possible expansions the client can do.

    The all_output dict is a IO param, returning info to caller.
    """

    # init
    instruments = sorted(SUPPORTED_MODES[telescope])

    # order the output. this list contains either key-names or comment lines
    if 'order' not in all_output: all_output['order'] = []

    all_output['order'] += [
        ('expansion_lines', 'Batch expansions enabled in the client:'),
    ]
    all_output['expansion_lines'] = []

    # loop through all supported instruments and modes
    for instr in instruments:

        # open the instrument config file
        fname = pandeia_data_loc+f'/{telescope}/'+instr+'/config.json'
        cfg_tree = json.loads(read_full_file(fname))
        assert type(cfg_tree) == dict, 'Unexpected json layout in: '+fname
        strip_meta(cfg_tree)

        # loop over modes
        for mode in sorted(SUPPORTED_MODES[telescope][instr]):

            instr_mode = instr+'_'+mode
            if instr_mode in EXPANSIONS[telescope]:
                # mode config sub-dict
                mode_cfg = {}
                if 'mode_config' in cfg_tree:
                    mode_cfg = cfg_tree['mode_config'][mode]

                # For each possible expansion-menu-option an inst-mode may have (usually just 1
                # option, e.g. "Expand over filters", but there may be multiple expansion-menu-options
                # per inst-mode), we come up with all possible needed values.
                permitted_expansions = EXPANSIONS[telescope][instr_mode]
                expansion_out = {}
                menu_order = 0
                for expansion_dict in permitted_expansions:
                    # see which pars to vary
                    this_name = expansion_dict['name']
                    expansion_out[this_name] = {}
                    if 'match' in expansion_dict:
                        expansion_out[this_name]['match'] = expansion_dict['match']
                    expansion_out[this_name]['menu'] = menu_order
                    pars_to_vary = expansion_dict['pars']
                    expansion_out[this_name]['pars'] = pars_to_vary
                    expansion_out[this_name]['values_key'] = 'aperture'
                    expansion_out[this_name]['values'] = { '*': [] } # assume simple unconstrained case, '*' means "for any aperture"
#                   print('expansion: ' + str(expansion_out[this_name]))

                    # Interject some logic here to handle any all-comb situations (when a par to vary is a tuple of pars).
                    # In that case we actually explode the set of those pars out to the all-comb'ed version of full length.
                    # But we will replace that par in pars_to_vary with the 2 or 3 separate exploded pars.
                    exploded_vals = {}
                    for allcomb in pars_to_vary:
                        if isinstance(allcomb, tuple):
                            # We have found an all-combination tuple, so explode it out and cache
                            # the resulting values.  Assume at most 3 pars for this (for now).
                            assert len(allcomb) in (2,3), 'Unsupported all-comb: '+str(allcomb)
                            if len(allcomb) == 2:
                                ttt = allcomb + (None,) # grow to len=3
                                allcomb = ttt
                            # explode it
                            (par1name, par2name, par3name) = allcomb
                            assert par1name == 'aperture', 'Not yet handling all-combs where first par isnt aperture: '+par1name
                            exploded_raw = []
                            # collect par1vals
                            par1vals = get_possible_values_for(all_output, telescope, instr, mode, '*', par1name) # returns [None] for None par-name
                            # loop over par1vals
                            for p1 in par1vals:
                                # collect par2vals
                                par2vals = get_possible_values_for(all_output, telescope, instr, mode, p1, par2name)
                                # remove degenerate cases where there is only one possible constrained value
                                if type(expansion_out[this_name]['pars'][0]) == tuple and \
                                   len(expansion_out[this_name]['pars'][0]) == 2 and \
                                   len(par2vals) < 2:
                                    par2vals = []
                                # loop over par2vals
                                for p2 in par2vals:
                                    # collect par3vals
                                    par3vals = get_possible_values_for(all_output, telescope, instr, mode, p1, par3name)
                                    # remove degenerate cases where there is only one possible constrained value
                                    if type(expansion_out[this_name]['pars'][0]) == tuple and \
                                       len(expansion_out[this_name]['pars'][0]) == 3 and \
                                       len(par3vals) < 2:
                                        par3vals = []
                                    # loop over par3vals then append to exploded_raw list
                                    for p3 in par3vals:
                                        exploded_raw.append( [p1, p2, p3] )

                            exploded_columns = list(zip(*exploded_raw))
                            exploded_vals[par1name] = exploded_columns[0]
                            exploded_vals[par2name] = exploded_columns[1]
                            exploded_vals[par3name] = exploded_columns[2]

                    # flatten the pars_to_vary list since we have now exploded any if there were any
                    flat_pars_to_vary = []
                    for ppp in pars_to_vary:
                        if isinstance(ppp, tuple):
                            flat_pars_to_vary += list(ppp)
                        else:
                            flat_pars_to_vary.append(ppp)
                    pars_to_vary = flat_pars_to_vary                # overwrite
                    expansion_out[this_name]['pars'] = pars_to_vary # overwrite

                    # From here down, everything in pars_to_vary is "zippered" together (e.g. for 2 pars, not ALL combinations of both are used, only a "zipper" of them both)
                    for par_to_vary in pars_to_vary:
                        assert par_to_vary in ('aperture','disperser','filter','reference_wavelength'), \
                            'Not yet handling expansions which vary the following par or par-combo: %s' % str(par_to_vary)
                        # lookup the ref wave?
                        if par_to_vary == 'reference_wavelength':
                            # collect the ref wave values _based-on_ a previously-added set of values (e.g. filter)
                            poss_vals = []
                            if ('filter' in pars_to_vary and pars_to_vary.index('reference_wavelength') > pars_to_vary.index('filter')) or \
                               ('disperser' in pars_to_vary and pars_to_vary.index('reference_wavelength') > pars_to_vary.index('disperser')):
                                idx_of_prior_par = -1
                                idx_of_aper_par = -1 if 'aperture' not in pars_to_vary else pars_to_vary.index('aperture')
                                if 'filter' in pars_to_vary:
                                    idx_of_prior_par = pars_to_vary.index('filter')
                                elif 'disperser' in pars_to_vary:
                                    idx_of_prior_par = pars_to_vary.index('disperser')
                                else:
                                    raise ValueError('Not yet handling "prior par"s besides filter and disperser')
                                for i in range(len(expansion_out[this_name]['values']['*'])):
                                    aper_for_refwave = 'imager' # 'imager' is hard-coded if not overwritten next !
                                    if idx_of_aper_par > -1:
                                        aper_for_refwave = expansion_out[this_name]['values']['*'][i][idx_of_aper_par]
                                    prior_par_value = expansion_out[this_name]['values']['*'][i][idx_of_prior_par]
                                    assert prior_par_value in REFWAVE_DICT[instr][mode][aper_for_refwave] and \
                                           'wmidpoint' in REFWAVE_DICT[instr][mode][aper_for_refwave][prior_par_value], \
                                           'May be dealing with a complex refwave dict here for %s %s %s' % (instr, mode, aper_for_refwave)
                                    wmidpoint = REFWAVE_DICT[instr][mode][aper_for_refwave][prior_par_value]['wmidpoint']
                                    poss_vals.append(wmidpoint)
                            else:
                                raise ValueError('Cannot decide a ref wave in an expansion without a separate prior par to use for lookup')
                        # not a ref wave, then maybe just a filter or similar
                        else:
                            # grab the possible values for this par (assume simple version w/ no constraints)
                            if par_to_vary in exploded_vals:
                                poss_vals = exploded_vals[par_to_vary]
                            else:
                                poss_vals = get_possible_values_for(all_output, telescope, instr, mode, '*', par_to_vary)

                        if len(expansion_out[this_name]['values']['*']) == 0:
                            # first time through, so collect values for the first par listed in pars_to_vary; but in [ [,], [,] ] form
                            expansion_out[this_name]['values']['*'] = [[f,] for f in poss_vals]
                        else:
                            # we have already collected values from a previous par(s); pair these values to those
                            assert len(poss_vals) == len(expansion_out[this_name]['values']['*']), \
                                'Cannot pair up values; mismatched value-set lengths! Trying to zipper-add:\n'+ \
                                str(poss_vals)+'\nonto:\n'+str(expansion_out[this_name]['values']) # should never happen
                            for i in range(len(poss_vals)):
                                expansion_out[this_name]['values']['*'][i].append(poss_vals[i])

                    # sometimes, due to bloated constraints, we have needlessly bloated expansions, so consolidate if so
                    expansion_out[this_name] = consolidate_bloated_expansion_dict(expansion_out[this_name])

                    # move on to next permitted expansion
                    menu_order += 1

                def expansion_sorter(dict_key):
                    return expansion_out[dict_key]['menu']

                all_output['expansion_lines'].append(f'CFG_VALS.EXPANSIONS_{telescope}_{instr}_{mode} = {dict_to_str(expansion_out, key_sorter=expansion_sorter)};')
    # done

def generate_mappings(pandeia_data_loc, all_output):
    """ This is a stand-in until we see how this is likely to be used. """

    all_output['order'] += [ ('mapping_lines','Miscellaneous mappings of values to display strings for the client:') ]

    short_str_mappings = name_mapper(None) # returns mapping dict

    all_output['mapping_lines'] = [f'CFG_VALS.MISC_DISP_STR_MAP = {dict_to_str(short_str_mappings)};']


def make_string_replacements(input_string, replacements_dict):
    """ Convenience function.  Note this also occurs in pandeia_test/test_utilities/client/regen_lib """
    retval = input_string[:]
    for key in replacements_dict:
        retval = retval.replace(key, replacements_dict[key])
    return retval


# -------------------------------------------------------------------------------------------------

def write_config_values(telescope, all_output, pandeia_data_loc, output_dir, verbose):
    """ Dump the generated config values as JS to a .js file. """

    # write header of CFG_VALS file
    cfg_out_nm = f'{output_dir}/config_{telescope}.js'
    if os.path.exists(cfg_out_nm): raise RuntimeError('Will not overwrite: '+cfg_out_nm)
    cfg_out = open(cfg_out_nm, 'w')

    cfg_out.write('//'+80*'-'+'\n')
    cfg_out.write('// DO NOT EDIT!  This file is generated via "pandeia_data/devtools/generate_js.py"\n')
    cfg_out.write('// (copy this re-generated file into pandeia/ui/client/js, over the existing one)\n')
    cfg_out.write('// DO NOT EDIT!\n')
    cfg_out.write('//'+80*'-'+'\n') # -----------------------------------------------------------------------

    cfg_out.write('\nvar CFG_VALS = {};\n')

    cfg_out.write('\n// Creation data:\n')
    cfg_out.write("CFG_VALS.CREATED_time_asc = '"+time.asctime()+"';\n")
    cfg_out.write("CFG_VALS.CREATED_time_float = "+str(time.time())+';\n')

    # now dump by keys in all_output, as listed in its "order" list
    for key_tup in all_output['order']:
        assert key_tup[0] in all_output, 'Dev error: key found in "order" list but not found in dict?: %s' % key_tup[0]
        # write it out only if there are things to write
        if len(all_output[key_tup[0]]) > 0:
            # first see if it has a comment block to write out
            if len(key_tup) > 0 and key_tup[1]:
                cfg_out.write('\n//'+100*'-'+'\n//\n')
                cfg_out.write('// '+key_tup[1]+'\n//\n')
            else:
                cfg_out.write('\n//'+80*'-'+'\n')
            # it is a key (e.g. 'supported_lines') for a set of output lines, so dump them
            for line in all_output[key_tup[0]]:
                # write out each line but not if it is an empty item, unless it is required
                if 'UNFINISHED' in line or 'SUPPORTED' in line or not (line.endswith(' = [];') or line.endswith(' = {};')):
                    cfg_out.write(line+'\n')

    cfg_out.close()
    if verbose:
        print(f'\nGENERATED for {telescope}: {cfg_out_nm}')


def write_client_validation_data(telescope, all_output, pandeia_data_loc, output_dir):
    """ Dump the generated client-vlad-constraints as JS to a .js file. """

    # write header of the vlad constraints file
    js_out_nm = output_dir+f'/val_constraint_data_{telescope}.js'
    if os.path.exists(js_out_nm): raise RuntimeError('Will not overwrite: '+js_out_nm)
    js_out = open(js_out_nm, 'w')

    js_out.write('//'+80*'-'+'\n')
    js_out.write('// DO NOT EDIT!  This file is generated via "pandeia_data/devtools/generate_js.py"\n')
    js_out.write('// (copy this re-generated file into pandeia/ui/client/js, over the existing one)\n')
    js_out.write('// DO NOT EDIT!\n')
    js_out.write('//\n// "vlad" is an etc-colloquial shorthand for "validation"\n//\n')
    js_out.write('// See all documentation of constraint fields in val_constraints.js.\n')
    js_out.write('//'+80*'-'+'\n') # -----------------------------------------------------------------------

    # Go through the VLAD_CONSTRAINTS looking for any expansions that need to be made
    # So far supported: ${ALL_INST_MODES}

    vlad_constraints = val_constraint_data.VLAD_CONSTRAINTS_JWST
    if telescope == 'hst':
        vlad_constraints = val_constraint_data.VLAD_CONSTRAINTS_HST

    for vlad_name in vlad_constraints:
        vlad = vlad_constraints[vlad_name]
        assert 'trigger_ids' in vlad, 'CHANGE NEXT LINE!'
        if 'trigger_ids' in vlad and '${ALL_INST_MODES}' in str(vlad['trigger_ids']):
            inst_condition = [c for c in vlad['conditions'] if 'data_name' in c and '==' in c and c['data_name']=='s_instrument'][0]
            inst = inst_condition['==']
            all_inst_modes = [inst+'_'+m for m in sorted(SUPPORTED_MODES[telescope][inst])]
            new_trigger_ids = []
            for trigger_id in vlad['trigger_ids']:
                if '${ALL_INST_MODES}' in trigger_id:
                    new_trigger_ids += [trigger_id.replace('${ALL_INST_MODES}',im) for im in all_inst_modes]
                else:
                    new_trigger_ids.append(trigger_id)
            vlad_constraints[vlad_name]['trigger_ids'] = new_trigger_ids


    # !!! temporarily sort these vlads in the same order they were used/listed in the client
    # !!! before they were generated. this is to simplify testing and isolate any diffs on
    # !!! generated JS and in behavior.  later we will pull this sorter out and go alphabetical.
    tmp_sort_order = ["niriss_soss_ord2_sub96", "niriss_soss_sub_with_nis", "niriss_ami_sub_with_nis",
                      "nirspec_fs_irs2_w_full", "nircam_coro_w_filt_plus_nd", "nircam_wfgrism_grismc_orientation",
                      "miri_modes_w_unsupported_slow_bh7", "miri_modes_w_unsupported_slow_bh8", "miri_low_ngroups_vals"]
    tmp_sort_order_str = '__' + '__'.join(tmp_sort_order) + '__'
    # keeps the above-listed vlads in their original (tested!) order, but put the rest into a deterministic
    # alpha-based order that is the same on both py2 and py3. later rip out both tmp_sort_order and key_sorter!
    # !!! end temporary code

    vlad_blob = dict_to_str(vlad_constraints, indent=1,
                            key_sorter=lambda k: \
                            sorted(vlad_constraints, reverse=True).index(k) + \
                            max(0, tmp_sort_order_str.find('__'+k+'__')*1000)
                            )

    vlad_blob = unquote_js_varnames(vlad_blob)
    js_out.write('pandeia.cfg.vlad_constraints = '+vlad_blob+';\n')

    js_out.close()


def write_help_data(telescope, all_output, pandeia_data_loc, output_dir):
    """ Read then dump the help info as JS to a .js file. """

    # write header of the vlad constraints file
    js_out_nm = output_dir+f'/help_links_{telescope}.js'

    if os.path.exists(js_out_nm): raise RuntimeError('Will not overwrite: '+js_out_nm)
    js_out = open(js_out_nm, 'w')

    js_out.write('//'+80*'-'+'\n')
    js_out.write('// DO NOT EDIT!  This file is generated via "pandeia_data/devtools/generate_js.py"\n')
    js_out.write('// (copy this re-generated file into pandeia/ui/client/js, over the existing one)\n')
    js_out.write('// DO NOT EDIT!\n')
    js_out.write('//'+80*'-'+'\n') # -----------------------------------------------------------------------

    # read in the help data from JSON
    fname = pandeia_data_loc+f'/webapp/help_data_{telescope}.json'
    help_tree = json.loads(read_full_file(fname))
    assert type(help_tree) == dict, 'Unexpected json layout in: '+fname
    strip_meta(help_tree)

    # make any substitutions as needed
    repls = help_tree['substitutions']
    if 'comment' in repls:
       repls.pop('comment')
    out_tree = {
        'direct_links': {},
        'hover_text': {},
        'main_dox_link': make_string_replacements(help_tree['main_dox_link'], repls),
    }
    # copy help tags into out dict, substituting along the way
    for tag in [key for key in help_tree['direct_help_links'] if key != 'comment']:
        # handle expansions
        if '__IM__' in tag:
            # expand this tag over every instrument/mode combo in this telescope (eg help icon is in templated html)
            # sorted(SUPPORTED_MODES['jwst'])    for mode in sorted(SUPPORTED_MODES[telescope][instr])
            for inst in sorted(SUPPORTED_MODES[telescope]):
                for mode in sorted(SUPPORTED_MODES[telescope][inst]):
                    expandedtagname = tag.replace('__IM__',inst+'_'+mode)
                    out_tree['direct_links'][expandedtagname] = [make_string_replacements(item, repls) for item in help_tree['direct_help_links'][tag]]
        else:
            # straight up tag
            out_tree['direct_links'][tag] = [make_string_replacements(item, repls) for item in help_tree['direct_help_links'][tag]]

    # Dump to the file
    js_out.write('pandeia.cfg.help = '+dict_to_str(out_tree, indent=1)+';\n')

    # Now also pop in there 3 JS helper functions to open the help locations
    helper = """
function clicked_help_link(tag)
{
    // simple helper function, not specific to any ETC page, to track that a help link was clicked
    console.log('open_help for tag: '+tag); // goes to ES/kibana
    // tell google analytics too; use tag since likely more stable than the urls
    // plan to use "navigate" for clicks and "hover" for the future hover-help-text
    gtag('event', 'navigate', { 'event_category': 'context-sensitive-help', 'event_label': tag });
}

function enable_help_link(elem_id)
{
    console.log('Enabling help link: '+elem_id);
    // given the elem_id (same as the help tag name) grab the url info
    var url = 'https://%s-docs.stsci.edu'
    if (pandeia.cfg && pandeia.cfg.help && pandeia.cfg.help['direct_links'][elem_id] != undefined) {
        url = pandeia.cfg.help['direct_links'][elem_id][0];
    } else {
        console.log('Cannot find help data for tag: '+elem_id); // not an error
    }
    // give this element the attrs to be a proper link
    $('#'+elem_id).attr('href', url);
    $('#'+elem_id).attr('target', '_blank');
    // hook in an on-click for us
    $('#'+elem_id).on('click', function() { clicked_help_link(elem_id); });
}

function enable_static_help_links()
{
    var help_link_a_elements = $('a.etc_help_a'); // a jquery set object
    $.each(help_link_a_elements, function(idx) {
        var elem_id = help_link_a_elements[idx].id;
        enable_help_link(elem_id);
    });
}
""" % (telescope)
    js_out.write(helper)

    # Done
    js_out.close()


def usage():
    print(sys.argv[0]+' <jwst|hst> <path-to-pandeia_data> [-h (for help)] [-v (for verbose)]')
    sys.exit(0)


#
# main routine
#
if __name__=='__main__': # in case something else imports this file

    if '-h' in sys.argv or len(sys.argv) < 2 or sys.argv[1] not in ('hst','jwst'):
        usage()

    telescope = sys.argv[1]
    verbose = '-v' in sys.argv

    if len(sys.argv) > 2:
        data_dir = sys.argv[2]
    else:
        data_dir = os.environ['pandeia_refdata']
    pandeia_data_loc = os.path.realpath(data_dir)

    # all ref waves we might need
    REFWAVE_DICT = iou.get_full_refwave_dict(pandeia_data_loc,
                                             telescope,
                                             list(SUPPORTED_MODES[telescope]),
                                             SPEC_MODENAMES[telescope])

    # the output object
    all_output = {}
    all_output['CACHED'] = {} # stuff figured in one step that may be needed for a later step

    generate_cfg_basics(pandeia_data_loc, telescope, all_output, verbose)
    generate_expansions(pandeia_data_loc, telescope, all_output)
    generate_mappings(pandeia_data_loc, all_output)

    # write the files
    output_dir = os.getcwd()
    try:
        write_config_values(telescope, all_output, pandeia_data_loc, output_dir, verbose)
        write_client_validation_data(telescope, all_output, pandeia_data_loc, output_dir)
        write_help_data(telescope, all_output, pandeia_data_loc, output_dir)
    except RuntimeError as rte: # let all others fly
        print('ERROR: '+str(rte))
        sys.exit(1)

    # for debugging help, dump the CACHED lines
#   for key in sorted(all_output['CACHED']):
#       print('')
#       print(key, all_output['CACHED'][key])

    sys.exit(0)
