#!/usr/bin/env python
"""
If there is a new valid filter combination for a given aperture (or a change to the wavelength range covered by a filter,
or large changes to the instrument optical models) you will need to regenerate the PSFs to cover the entire wave range
(our PSF library functions will interpolate between those PSFs as needed)

What you will need:

    * A Python 3 Third Party environment
    * The pandeia_data/devtools/make_psf2.py script

From STScI's Astroconda channel (or pip, which seems to be updated first):

    * A copy of WebbPSF, which uses instrument parameters to generate PSFs for an instrument, mode, wavelength, and
      position of source.
    * A copy of POPPY, the optical physics propagation toolkit WebbPSF uses to simulate the optical elements.
    * The latest WebbPSF data delivery (either local, or on the STScI servers)
    * This code added to your .bashrc (or at least in the shell you're using):
       `export WEBBPSF_PATH="/grp/jwst/ote/webbpsf-data" where that path is to the WebbPSF data. Astroconda should set
       this for you.

make_psf2.py is intended to be run from the console as `python make_psf2.py n_threads instrument,aperture1`
where the instrument and aperture arguments are optional. Examples:

 * `python make_psf2.py 4` Will run ALL JWST instrument PSFs for ALL apertures, using 4 parallel threads

 * `python make_psf2.py 3 niriss` Will run ALL NIRISS PSFs for all its apertures, using 3 parallel threads

 * `python make_psf2.py 1 miri,ch1,ch4` Will make the MIRI ch1 and ch4 PSFs using only one thread

 * `python make_psf2.py 8 nirspec,shutter,ifu nircam,maskswb` Will make NIRSpec Shutter and IFU, and NIRCam maskswb PSFs
   using 8 parallel threads.

 * `python make_psf2.py 2 wfirstimager,grism,any` Will make the WFIRST Imager PSFs for the grism and imager. You MUST
   specify instrument and all apertures for WFIRST (otherwise the convenience functions try to look them up in jwst folders)

Most of the code inside is to configure the WebbPSF instrument objects, and setting up special exceptions like turning
off WebbPSF's internal handling of the NIRCam occulting bar offsets because we want to set r and theta ourselves.

Settings for PSFs (wavelength ranges, pixel scales, oversampling, whether to include the measured wavefront errors, etc)
are read from config.json and psf.json.

Using more cores tends to overheat a laptop and slow it down, so there's really very little difference between setting
n_threads to anything above 3. As of October 2017, the process of generating the entire suite of PSFs with n_threads = 4
took 5 hours, and generated 2.5 GB of files. Accordingly, you will want to upload the PSFs to pandeia_data in small chunks
(maybe 500 at a time).

One unresolved issue with this: git stores every previous commit on the hard drive (for easy access), so as we remove and
replace files, pandeia_data's hidden history files are going to grow immensely. Update PSFs with care.

Program flow:
1. __main__ creates a PSFInstrument class object of the right type
2. __init__ (for each instrument) creates a WebbPSF instrument with the requested apertures identified, then hands off to
   the generic __init__.
3. __init__ (generic) loads that config.json and psf.json file, and runs _load_config
4. _load config takes the generic instrument configuration out of the three different areas of the psf.json file
   (instrument, options, and general) and applies them to either the PSFInstrument object or the WebbPSF instrument.
5. for each requested aperture, __main__ runs the .create_psf() function of the PSFInstrument class
6. create_psf() sets up aperture-specific masks, elements, wavelength ranges, wavelength grids, and (if applicable) hands
   off to coronagraphy-specific functions to set up all the offsets. It hands off to setup_psf() (or setup_coro_psf,
   where applicable.)
7. setup_psf (or setup_coro_psf) generates one PSF setup for every wavelength (and offset) and adds them to a list.
8. _split_multi takes the list of all PSF setups to generate, calculates the pupil throughput, sets up the requested
   number of multiprocessing cores, and then splits up the list to each multiprocessing job queue.
9. (in each thread) _psf_runner takes the job list it's been given, and runs gen_psf on each PSF setup
10. gen_psf takes all the information bundled up into the PSF setup and generates the actual correctly-named FITS file.

"""
import webbpsf as wp
import numpy as np
import os
import multiprocessing as mp
from sys import argv
import json

from astropy.io import fits

from scipy import special as sp
from scipy import optimize as op

from pandeia_test import data_util

COMPRESSION = False
MAKE_PSF_VERSION = 4
#COMPRESSION = 65536

# The maximum FOV size is 512 lambda/D (according to M. Perrin, priv. comm.)
MAX_FOV_SIZE = 512 #arcsec

"""
Except on Python 3, there is no reliable way to calculate the pupil throughput
and THEN begin iterating. To avoid each subprocess generating PSFs with a different
pupil_throughput keyword, you should NOT run this on python 2 with more than 1 core.
"""

class PSFInstrument():
    """
    Parent class for all instruments.
    If you get a Zernicke coefficients error, you need to copy the measured wavefront errors file,
    /grp/jwst/ote/si_zernikes_isim_cv3.fits, to the webbpsf-data folder.

    Parameters
    ----------

    Attributes
    ----------
        data_dir: string
            The location of the Pandeia Reference Data
        config: dict
            The instrument configuration file, as a dictionary
        instrument: WebbPSF Instrument
            A JWSTInstrument class from WebbPSF

        aperture: string
            The aperture for which PSFs are being requested.
        aperture_name: string
            The aperture name to use in the fits file and its filename (different only for NIRSpec)
        coronagraph_source_offsets: list
            r, theta pairs of source offset positions for use with coronagraphic modes
        fov_pixels: int
            Size of one dimension of the (square) FOVs in pixels, before oversampling.
        include_si_wfe: book
            Indicates whether to load the measured wavefront errors (as opposed
            to the predicted ones)
        minwave: float
            The minimum wavelength defined in WebbPSF, for which PSFs can be generated.
            Overridden by NIRISS and NIRCAM
        maxwave: float
            The maximum wavelength defined in WebbPSF for which PSFs can be generated.
            Overridden by NIRISS and NIRCAM
        n_cores: int
            The number of threads to use in parallel to generate PSFs.
        nstep: int
            The step skip size to use when cutting down the master wavelength grid to a
            reasonable number of points.
        oversample: int
            The number of points to oversample the PSF. Overridden by NIRSpec
        trim_fov_pixels: int
            Trim the PSF to just this many pixels, centered. Used only by NIRCAM

    Methods
    -------
    create_psfs:
        Sets up lists of all the PSFs that need to be made for that instrument/aperture
        combination. Overridden by NIRCAM for maskswb and masklwb because the bar psfs
        require a lot of special treatment.
    _master_wave:
        Creates the wavelength grid that all instruments and apertures are snapped into
    _get_wave_range:
        Given an aperture and filter(s), get the wavelength range specified in the config
        and then find what grid values to snap to
    _get_wave_elements:
        Given an aperture, find all the filters and dispersers that are meant to be used
        with it
    _split_multi:
        Divide up the list of PSFs to generate between some number of threads
    genpsf:
        Accepts one PSF dictionary, outputs one generated PSF in a .fits file
    _psf_runner:
        Used by _split_multi to run an entire (sub-)list of PSFs through genpsf


    """
    def __init__(self):
        self.data_dir = os.environ['pandeia_refdata']
        config_file = open('{0:}/{1:}/{2:}/config.json'.format(self.data_dir,self.telescope,self.insname))
        self.config = json.load(config_file)

        psf_file = open('{0:}/{1:}/{2:}/psf.json'.format(self.data_dir,self.telescope,self.insname))
        self.psfconfig = json.load(psf_file)

        #self.outfile = open('{}_psfparam.csv'.format(self.insname),'w')

        # apply PSF configuration from the file
        self._load_config()

        self._master_wave()

    def _load_config(self):
        """
        Load PSF configuration from a file.

        The general structure is a json file, with three major categories:

        "instrument" contains global instrument properties to be loaded directly into the
            WebbPSF instrument.
        "general" contains global instrument properties that should live in the Class object.
        "aperture" contains aperture-specific properties, and is itself divided into
            instrument and general sub-categories.
        There are additional attributes that do not fall under those umbrella headings.
        """

        # load and set global instrument configuration
        for item in self.psfconfig['instrument']:
            setattr(self.instrument,item,self.psfconfig['instrument'][item])
        for item in self.psfconfig['options']:
            self.instrument.options[item] = self.psfconfig['options'][item]
        for item in self.psfconfig['general']:
            setattr(self,item,self.psfconfig['general'][item])

        # load aperture-specific configuration
        for item in self.psfconfig['apertures'][self.aperture]['instrument']:
            setattr(self.instrument,item,self.psfconfig['apertures'][self.aperture]['instrument'][item])
        for item in self.psfconfig['apertures'][self.aperture]['general']:
            setattr(self,item,self.psfconfig['apertures'][self.aperture]['general'][item])

        # detector scattering is special
        if 'detector_scat' in self.psfconfig:
            for item in self.psfconfig['detector_scat']:
                setattr(self,item,self.psfconfig['detector_scat'][item])
            self.detector_scat = True

        # pixel scale is special
        self.instrument.pixelscale = self.config['aperture_config'][self.aperture]['pix']

        # minwave and maxwave are also special
        if 'minwave' in self.psfconfig['apertures'][self.aperture]:
            if self.psfconfig['apertures'][self.aperture]['minwave'] == 'short_min':
                self.minwave = self.instrument.SHORT_WAVELENGTH_MIN*1e6
            elif self.psfconfig['apertures'][self.aperture]['minwave'] == 'long_min':
                self.minwave = self.instrument.LONG_WAVELENGTH_MIN*1e6
            else:
                raise ValueError('Unable to read wavelengths')
        if 'maxwave' in self.psfconfig['apertures'][self.aperture]:
            if self.psfconfig['apertures'][self.aperture]['maxwave'] == 'short_max':
                self.maxwave = self.instrument.SHORT_WAVELENGTH_MAX*1e6
            elif self.psfconfig['apertures'][self.aperture]['maxwave'] == 'long_max':
                self.maxwave = self.instrument.LONG_WAVELENGTH_MAX*1e6
            else:
                raise ValueError('Unable to read wavelengths')

    def _detector_scat(self,wave):
        '''
        The loss due to detector scattering in the
        MIRI imaging detector. A function of wavelength, returns a float between 0 and 1
        The detector may scatter light to large distances, leading to an effective
        loss of throughput in sources (the background is not affected). This is
        an issue for MIRI, but not currently for the other instruments
        (see Glasse et al. 2015). This function will decrease the PSF throughput by
        a factor.
        '''
        f_det = 1. - self.a_det * np.exp(-(self.tau1 + self.tau2) * (wave / self.wave_0)**2.)
        return f_det


    def create_psfs(self,n_cores):
        """
        Set up PSFs by setting up all the names, wavelengths, and offsets

        Parameters
        ----------
        n_cores: int
            The number of parallel generation instances to run
        """

        if not hasattr(self,'offsets'):
            self.offsets = [(0.0,0.0)]

        if hasattr(self,'elements'):
            for element in self.elements:
                self.aperture_name = '{0:}{1:}'.format(self.aperture,element['name'])
                waveelements = element['waveelements']
                wavelengths = self._get_wave_range_fixedgrid(waveelements)
                self.instrument.pupil_mask = element['pupil_mask']
                self.instrument.image_mask = element['image_mask']
                self.offsets = element['offsets']
                self.setup_psfs(n_cores,wavelengths)

        else:
            waveelements = self._get_wave_elements()
            print(waveelements)
            wavelengths = self._get_wave_range_fixedgrid(waveelements)
            try:
                self.aperture_name # the only aperture that should already have this is nirspec
            except AttributeError:
                self.aperture_name = self.aperture
            self.setup_psfs(n_cores,wavelengths)


    def setup_psfs(self,n_cores,wavelengths):
        """
        Create each PSF object and then send them to the system that splits
        them off into parallel generation.

        Parameters
        ----------
        n_cores: int
            The number of parallel generation instances to run
        wavelengths: ndarray
            The wavelengths at which to generate PSFs
        """
        psfs = []

        # reverse the order so that the unocculted PSF (if it's coronagraphy) is used to determine pupil throughput
        for dx,dy in self.offsets[::-1]:

            for wave in wavelengths:
                if np.sqrt(dx**2 + dy**2) > 0.001:
                    # Basename + aperture + wavelength + (if non-trivial) x and y offsetfrom on-axis as defined in WebbPSF
                    longname = '{0:}_{1:}_{2:.4f}_{3:.3f}_{4:.3f}.fits'.format(self.insname,self.aperture_name,wave,dx,dy)
                else:
                    longname = '{0:}_{1:}_{2:.4f}.fits'.format(self.insname,self.aperture_name,wave)

                # do filename check here for speed reasons
                if os.path.isfile(longname):
                    print("{0:} already exists. skipping...".format(longname))
                else:
                    r = np.sqrt(dx**2+dy**2)
                    theta = np.degrees(np.arctan2(dy,dx))
                    psf = {'psf':self.instrument, 'wave':wave*1e-6, 'filename':longname, 'optimal_offset_x': dx, 'optimal_offset_y': dy, 'source_offset_r': r, 'source_offset_theta': theta}
                    psfs.append(psf)
        if len(psfs) > 0:
            self._split_multi(n_cores,psfs)

    def _master_wave(self):
        """
        Calculate a wavelength grid for all PSFs to be sampled from.

        Keywords
        --------
        wmin: float
            the minimum wavelength for the master grid
        wmax: float
            the maximum wavelength for the master grid
        n: int
            the number of steps in the master grid

        Returns
        -------
        wvals: ndarray
            the master log-normal wavelength grid
        """

        n = 2000
        wmin = 0.3
        wmax = 34.0
        a = (n -1)/(np.log(wmax) - np.log(wmin))
        b = n - (a * np.log(wmax))

        arr = np.arange(n)+1
        self.masterwave = np.exp((arr - b)/a)

    def _get_wave_range_fixedn(self,waveelements):
        if self.aperture in ['maskswb', 'masklwb']:
            wrange = self.config['range'][aperture][waveelements[0]]
            self.genmin = wrange['wmin']
            self.genmax = wrange['wmax']
            self.gennum = int(10.0 * (self.genmax - self.genmin))
        nw = float(self.gennum)
        waves = (np.arange(nw) / (nw - 1)) ** 2. * (self.genmax - self.genmin) + self.genmin

        return waves

    def _get_wave_range_fixedgrid(self, waveelements):
        """
        Identify the required wavelengths for an actual calculation

        Keywords
        --------
        config: dict
            json dictionary of instrument configuration
        aperture: str
            the requested aperture to get range
        nstep: int
            the density of sampling into the master wavelength grid

        Returns
        -------
        waves: ndarray
            the wavelength points needed to span the wavelength range at the specified scaling
        """

        minwave = []
        maxwave = []

        if len(waveelements) != 0:
            for waveelement in waveelements:
                try:
                    maxwave.append(self.config['range'][self.aperture][waveelement]['wmax'])
                    minwave.append(self.config['range'][self.aperture][waveelement]['wmin'])
                except KeyError:
                    try:
                        for item in self.config['range'][self.aperture][waveelement]:
                            maxwave.append(self.config['range'][self.aperture][waveelement][item]['wmax'])
                            minwave.append(self.config['range'][self.aperture][waveelement][item]['wmin'])
                    except KeyError:
                        try:
                            maxwave.append(self.config['range'][self.aperture]['wmax'])
                            minwave.append(self.config['range'][self.aperture]['wmin'])
                        except KeyError:
                            print('   Wavelengths for {0:}/{1:} are missing!'.format(self.aperture, waveelement))
        else:
            try:
                maxwave.append(self.config['range'][self.aperture]['wmax'])
                minwave.append(self.config['range'][self.aperture]['wmin'])
            except KeyError:
                print('   Wavelengths for {0:} are missing!'.format(self.aperture))

        minwave = np.min(minwave)
        maxwave = np.max(maxwave)

        if minwave < self.minwave:
            print("Requested wavelength ({0:.4f}) below minimum defined ({1:.4f}) for {2:}".format(minwave,self.minwave,self.insname))
        if maxwave > self.maxwave:
            print("Requested wavelength ({0:.4f}) above maximum defined ({1:.4f}) for {2:}".format(maxwave,self.maxwave,self.insname))

        # cut down to just the step size steps
        mwave = [self.masterwave[x] for x in range(len(self.masterwave)) if x%self.nstep == 0]
        # grab the last wavelength BEFORE they become greater than minwave
        mindex = [k for k in range(len(mwave)) if mwave[k] <= minwave][-1]
        # grab the first wavelength AFTER they become greater than maxwave
        maxdex = [k for k in range(len(mwave)) if mwave[k] >= maxwave][1]

        waves = np.asarray(mwave[mindex:maxdex])

        # protect against wavelengths not supported by WebbPSF
        badmin = [k for k in range(len(waves)) if waves[k] < self.minwave]
        badmax = [k for k in range(len(waves)) if waves[k] > self.maxwave]
        waves[badmin] = self.minwave
        waves[badmax] = self.maxwave
        # and now protect against having MULTIPLES of the above.
        waves = np.unique(waves)

        return waves

    def _get_wave_elements(self):
        '''
        Given a json configuration dictionary and an aperture, find the filters
        and dispersers supported for that aperture.

        Returns
        -------
        elements: list
            a list of all filters and dispersers defined for that wavelength
        '''
        elements = []
        data_instrument = data_util.InstrumentConfiguration(telescope=self.telescope, instrument=self.insname)
        for mode in data_instrument.modes:
            data_instrument.set_mode(mode)
            data_instrument.webapp = True
            if self.aperture in data_instrument.apertures:
                data_instrument.set_aperture(self.aperture)
                print(mode, self.aperture, data_instrument.filters)
                elements.extend(data_instrument.dispersers)
                elements.extend(data_instrument.filters)

        elements = list(set(elements))

        # miri has a 'None' filter
        if None in elements:
            elements.remove(None)

        # niriss soss has one "disperser" for each order
        if hasattr(self,'special_elements'):
            elements.extend(self.special_elements)

        # We don't need a separate PSF library for the neutral density spots.
        elements = [element for element in elements if '_nd' not in element]

        return elements

    def _get_pupil_throughput(self,inputs, distortion=False):
        """
        Function to get the pupil throughput using a very wide aperture and the
        lowest wavelength requested for the filter combination.

        Note that the pupil loss is constant as a function of wavelength, so we
        only need to do this once per psf series.
        """

        source_offset_theta_rad = np.radians(inputs['source_offset_theta'])
        # Round to nearest integer subsampled pixel to avoid interpolations.
        # in webbpsf, when theta is 0, r is aligned with the +Y axis and theta increases CCW from there
        dx = inputs['source_offset_r'] * np.sin(source_offset_theta_rad)
        dy = inputs['source_offset_r'] * np.cos(source_offset_theta_rad)
        dx_pix = dx / (inputs['psf'].pixelscale / self.oversample)
        dy_pix = dy / (inputs['psf'].pixelscale / self.oversample)
        dx_pix_round = int(np.rint(dx_pix))
        dy_pix_round = int(np.rint(dy_pix))
        source_offset_r_pix = np.sqrt(dx_pix_round**2 + dy_pix_round**2)
        source_offset_theta = np.degrees(np.arctan2(dx_pix_round,dy_pix_round))

        inputs['psf'].options['source_offset_r'] = source_offset_r_pix * (inputs['psf'].pixelscale / self.oversample)
        inputs['psf'].options['source_offset_theta'] = source_offset_theta

        fov_arcsec = MAX_FOV_SIZE * inputs['wave']/self.telescope_mirror * 206265.0
        psf_full_fov = inputs['psf'].calc_psf(oversample=self.oversample,
                                  fov_arcsec=fov_arcsec,
                                  monochromatic=inputs['wave'],
                                  add_distortion=distortion)

        self.pupil_throughput = np.sum(psf_full_fov[0].data)

    def _split_multi(self,n_cores,psfs):
        """
        Split up the calculations into n_cores parallel threads

        Keywords
        --------
        n_cores: int
            The number of threads to split off into
        psfs: list
            A list of all the dictionaries setting up PSFs
        """
        self._get_pupil_throughput(psfs[0])

        #print('{}: {}'.format(psfs[0]['filename'],self.pupil_throughput))

        ctx = mp.get_context('forkserver')

        # now we have a list of all the PSFs we want to generate.
        processes = []
        for x in range(n_cores):
            to_generate = [psfs[k] for k in range(len(psfs)) if k%n_cores == x]
            processes.append(ctx.Process(target=self._psf_runner, args=([to_generate])))
            processes[-1].start()
        processes[-1].join() # block further execution until the last process finishes.

    def gen_psf(self, inputs, distortion=False):
        """
        Method to calculate a suite (range of wavelengths) of monochromatic PSFs
        with a format appropriate for use with Pandeia.

        Keywords
        --------
        inputs: dict
            A dictionary of all (changing) information necessary to generate the PSF.
        """

        wave = inputs['wave']
        instrument = inputs['psf']
        source_offset_r = inputs['source_offset_r']
        source_offset_theta = inputs['source_offset_theta']
        filename = inputs['filename'].lower()

        if os.path.isfile(filename):
            print("{0:} already exists. skipping...".format(filename))
            return
        else:
            print("Creating {0:}...".format(filename))

        # If off-axis, return center to origin
        if source_offset_r > 0.0001:
            source_offset_theta_rad = np.radians(source_offset_theta)
            # Round to nearest integer subsampled pixel to avoid interpolations.
            # in webbpsf, when theta is 0, r is aligned with the +Y axis and theta increases CCW from there
            dx = source_offset_r * np.sin(source_offset_theta_rad)
            dy = source_offset_r * np.cos(source_offset_theta_rad)
            dx_pix = dx / (instrument.pixelscale / self.oversample)
            dy_pix = dy / (instrument.pixelscale / self.oversample)
            dx_pix_round = int(np.rint(dx_pix))
            dy_pix_round = int(np.rint(dy_pix))
            source_offset_r_pix = np.sqrt(dx_pix_round**2 + dy_pix_round**2)
            source_offset_theta = np.degrees(np.arctan2(dx_pix_round,dy_pix_round))

            instrument.options['source_offset_r'] = source_offset_r_pix * (instrument.pixelscale / self.oversample)
            instrument.options['source_offset_theta'] = source_offset_theta
            #print(instrument.options['source_offset_r'],instrument.options['source_offset_theta'],source_offset_r_pix,"actual: ",dx_pix_round,dy_pix_round,"intended: ",dx_pix,dy_pix)
            dmax = np.max([np.abs(dx_pix_round), np.abs(dy_pix_round)])
            #instrument.options['fov_pixels']=self.fov_pixels
            instrument.calc_psf(oversample=self.oversample, fov_pixels=self.fov_pixels + 2 * dmax/3.0, monochromatic=wave,
                               outfile=filename, overwrite = True, add_distortion=distortion)

            # WebbPSF can make multiple planes, if requested:
            # 0: oversampled PSF
            # 1: non-oversampled PSF
            # 2: oversampled, distorted PSF (if add_distortion=True)
            # 3: non-oversampled, distorted PSF (if add_distortion=True)

            # Distortion effects mean the plane we want is #2. Without distortion,
            # the plane we want to save is #0.
            inpsf = fits.open(filename)
            if distortion:
                psf = fits.HDUList(fits.PrimaryHDU(data=inpsf[1].data,header=inpsf[1].header))
            else:
                psf = fits.HDUList(fits.PrimaryHDU(data=inpsf[0].data,header=inpsf[0].header))

            psf[0].data = np.roll(psf[0].data, dx_pix_round, axis=1)
            psf[0].data = np.roll(psf[0].data, -dy_pix_round, axis=0)
            psf[0].data = psf[0].data[dmax:(self.fov_pixels * self.oversample + dmax),
                                      dmax:(self.fov_pixels * self.oversample + dmax)]
            if hasattr(self,'trim_fov_pixels') and self.fov_pixels is not None:
                trim_amount = int(self.oversample * (self.fov_pixels - self.trim_fov_pixels) / 2)
                psf[0].data = psf[0].data[trim_amount:-trim_amount, trim_amount:-trim_amount]
            inpsf.close()
            os.remove(filename)
            psf.writeto(filename)
        else:
            instrument.options['source_offset_r'] = 0.
            instrument.options['source_offset_theta'] = 0.
            instrument.calc_psf(oversample=self.oversample, fov_pixels=self.fov_pixels, monochromatic=wave,
                               outfile=filename,overwrite = True, add_distortion=distortion)
            inpsf = fits.open(filename)
            if distortion:
                psf = fits.HDUList(fits.PrimaryHDU(data=inpsf[1].data,header=inpsf[1].header))
            else:
                psf = fits.HDUList(fits.PrimaryHDU(data=inpsf[0].data,header=inpsf[0].header))
            inpsf.close()
            os.remove(filename)
            psf.writeto(filename)

        # Add mode keyword to header
        hdulist = fits.open(filename,mode='update')
        prihdr = hdulist[0].header
        prihdr['MAKE_PSF'] = (MAKE_PSF_VERSION, 'Version of make_psf used to generate these files')
        prihdr['PUP_THRU'] = (self.pupil_throughput, 'The relative throughput of the pupil')
        prihdr['APERTURE'] = (self.aperture_name, 'The observing aperture within the instrument FOV')
        prihdr['OFFSET_R'] = (instrument.options['source_offset_r'], 'The radial distance in arcsec to the axis (snapped to grid)')
        prihdr['OFFSET_T'] = (instrument.options['source_offset_theta'], 'The source angle in degrees (snapped to grid)')
        if 'optimal_offset_x' in inputs and 'optimal_offset_y' in inputs:
            prihdr['OPTOFF_X'] = (inputs['optimal_offset_x'], 'x offset relative to optimal FOV position')
            prihdr['OPTOFF_Y'] = (inputs['optimal_offset_y'], 'y offset relative to optimal FOV position')

        if hasattr(self,'rot90'):
            hdulist[0].data = np.rot90(hdulist[0].data)

        if hasattr(self,'detector_scat') and distortion == False:
            hdulist[0].data *= self._detector_scat(wave*1e6)
            print("Applied {}".format(self._detector_scat(wave*1e6)))

        hdulist[0].data = np.flipud(hdulist[0].data)

        if COMPRESSION:
            comphdu = fits.CompImageHDU(hdulist[0].data,prihdr,compression_type='GZIP_1',quantize_level=COMPRESSION)
            hdulist.close()
            comphdu.writeto(filename, overwrite=True)
            comphdu.close()
        else:
            hdulist.writeto(filename, overwrite=True)
            hdulist.close()

    def _psf_runner(self,psfs):
        for psf in psfs:
            self.gen_psf(psf)


class JWSTInstrument(PSFInstrument):
    def __init__(self, aperture):
        self.telescope_mirror = 6.5 #meters
        self.telescope = 'jwst'
        self.aperture = aperture
        PSFInstrument.__init__(self)


class RomanInstrument(PSFInstrument):
    def __init__(self, aperture):
        self.telescope_mirror = 2.4 #meters
        self.telescope = 'roman'
        self.aperture = aperture
        PSFInstrument.__init__(self)

    def _load_config(self):
        """
        Load PSF configuration from a file.

        The general structure is a json file, with three major categories:

        "instrument" contains global instrument properties to be loaded directly into the
            WebbPSF instrument.
        "general" contains global instrument properties that should live in the Class object.
        "aperture" contains aperture-specific properties, and is itself divided into
            instrument and general sub-categories.
        There are additional attributes that do not fall under those umbrella headings.
        """

        # load and set global instrument configuration
        for item in self.psfconfig['instrument']:
            setattr(self.instrument,item,self.psfconfig['instrument'][item])
        for item in self.psfconfig['options']:
            self.instrument.options[item] = self.psfconfig['options'][item]
        for item in self.psfconfig['general']:
            setattr(self,item,self.psfconfig['general'][item])

        # load aperture-specific configuration
        # find pupil filename:
        try:
            pupil_mask = self.psfconfig['apertures'][self.aperture]['instrument']['pupil_mask']
        except KeyError:
            pass
        else:
            self.instrument.unlock_pupil()
            pupil_filename = self.instrument._pupil_controller.pupil_file_formatters[pupil_mask]
            basepath = self.instrument._pupil_controller._pupil_basepath
            pupil_path = os.path.join(basepath, pupil_filename.format('SCA_1'))
            self.instrument.lock_pupil(pupil_path)
        for item in self.psfconfig['apertures'][self.aperture]['general']:
            setattr(self,item,self.psfconfig['apertures'][self.aperture]['general'][item])

        # detector scattering is special
        if 'detector_scat' in self.psfconfig:
            for item in self.psfconfig['detector_scat']:
                setattr(self,item,self.psfconfig['detector_scat'][item])
            self.detector_scat = True

        # pixel scale is special
        self.instrument.pixelscale = self.config['aperture_config'][self.aperture]['pix']

        # minwave and maxwave are also special
        if 'minwave' in self.psfconfig['apertures'][self.aperture]:
            if self.psfconfig['apertures'][self.aperture]['minwave'] == 'short_min':
                self.minwave = self.instrument.SHORT_WAVELENGTH_MIN*1e6
            elif self.psfconfig['apertures'][self.aperture]['minwave'] == 'long_min':
                self.minwave = self.instrument.LONG_WAVELENGTH_MIN*1e6
            else:
                raise ValueError('Unable to read wavelengths')
        if 'maxwave' in self.psfconfig['apertures'][self.aperture]:
            if self.psfconfig['apertures'][self.aperture]['maxwave'] == 'short_max':
                self.maxwave = self.instrument.SHORT_WAVELENGTH_MAX*1e6
            elif self.psfconfig['apertures'][self.aperture]['maxwave'] == 'long_max':
                self.maxwave = self.instrument.LONG_WAVELENGTH_MAX*1e6
            else:
                raise ValueError('Unable to read wavelengths')

    def create_psfs(self,n_cores):
        """
        Set up PSFs by setting up all the names, wavelengths, and offsets

        Parameters
        ----------
        n_cores: int
            The number of parallel generation instances to run
        """

        if not hasattr(self,'offsets'):
            self.offsets = [(0.0,0.0)]

        if hasattr(self,'elements'):
            for element in self.elements:
                self.aperture_name = '{0:}{1:}'.format(self.aperture,element['name'])
                waveelements = element['waveelements']
                  #self.genmin,self.genmax = inst.waverange(waveelements=waveelements)
                wavelengths = self._get_wave_range_fixedgrid(waveelements)
                self.instrument.unlock_pupil()
                # find pupil filename:
                pupil_filename = self.instrument._pupil_controller.pupil_file_formatters[element['pupil_mask']]
                basepath = self.instrument._pupil_controller._pupil_basepath
                pupil_path = os.path.join(basepath, pupil_filename.format('SCA_1'))
                self.instrument.lock_pupil(pupil_path)
                self.offsets = element['offsets']
                self.setup_psfs(n_cores,wavelengths)

        else:
            waveelements = self._get_wave_elements()
            wavelengths = self._get_wave_range_fixedgrid(waveelements)
            try:
                self.aperture_name # the only aperture that should already have this is nirspec
            except AttributeError:
                self.aperture_name = self.aperture
            self.setup_psfs(n_cores,wavelengths)



class miri(JWSTInstrument):

    def __init__(self, aperture):
        """
        Set up a MIRI instrument and aperture, and MIRI-specific features

        Methods
        -------
        create_psfs:
            Override the default version so that FQPM PSFs can be handled differently
        create_coro_psfs:
            Special routine for handling the FQPM occulters
        _detector_scat
            MIRI is uniquely prone to scattering within the detector; this models that
            blurring effect as a throughput loss as a function of wavelength.
        """

        self.instrument = wp.MIRI()
        self.insname = 'miri'

        # load configuration
        super().__init__(aperture)

    def create_psfs(self,n_cores):
        if 'fqpm' in self.aperture:
            psfs = self.create_coro_psfs(n_cores)
        else:
            psfs = PSFInstrument.create_psfs(self,n_cores)
        return

    def create_coro_psfs(self,n_cores):
        """
        Set up PSFs by setting up all the names, wavelengths, and offsets

        Parameters
        ----------
        n_cores: int
            The number of parallel generation instances to run
        """

        if hasattr(self,'elements'):
            for element in self.elements:
                self.aperture_name = '{0:}{1:}'.format(self.aperture,element['name'])
                waveelements = element['waveelements']
                inst = data_util.InstrumentConfiguration(self.insname,self.telescope)
                inst.set_mode(inst.modes[0])
                inst.set_aperture(self.aperture)
                #self.genmin,self.genmax = inst.waverange(waveelements=waveelements)
                wavelengths = self._get_wave_range_fixedgrid(waveelements)
                self.instrument.pupil_mask = element['pupil_mask']
                self.instrument.image_mask = element['image_mask']
                self.offsets = element['offsets']
                self.setup_coro_psfs(n_cores,wavelengths)

        else:
            waveelements = self._get_wave_elements()
            wavelengths = self._get_wave_range_fixedgrid(waveelements)
            try:
                self.aperture_name # the only aperture that should already have this is nirspec
            except AttributeError:
                self.aperture_name = self.aperture
            self.setup_coro_psfs(n_cores,wavelengths)

    def setup_coro_psfs(self,n_cores,wavelengths):
        """
        Set up coronagraphy PSF sets for 2D patterns.

        For MIRI, we set up a triangle of points that can be reflected across
        eight axes of symmetry. We also need to rotate the entire coordinate
        system to the 4QPM axes.

        Parameters
        ----------
        n_cores: int
            The number of parallel generation instances to run
        wavelengths: list
            The wavelengths at which PSFs should be generated.
        """

        # the 4QPM mask is rotated -5 degrees w.r.t. the webbpsf coordinate system
        # exact value taken from WebbPSF 0.7.0
        bar_angle = 94.4497

        # build a triangle
        offsets = []
        for x in range(len(self.offsets)):
            for y in range(x+1):
                offsets.append([self.offsets[x],self.offsets[y]])

        psfs = []


        self.coronagraph_source_offsets = []
        # now loop through the offsets for positions around the optimal position
        # reverse the order so that the unocculted PSF is used to determine pupil throughput
        for dx,dy in offsets[::-1]:
            # get the r and theta to feed into webbpsf
            r_img = np.sqrt(dx**2 + dy**2)
            theta_img = np.degrees(np.arctan2(dy, dx)) - bar_angle
            self.coronagraph_source_offsets.append([r_img,theta_img,dx,dy])

        ## This code can be used, when uncommented, to overlay the PSF locations
        ## on a plot of the obscuration element.
        #lwb = wp.optics.MIRI_BandLimitedCoron(name='FQPM1065')
        #lwb.display(grid_size=10, colorbar=False, title=element)
        #for r_img,theta_img,dx,dy in self.coronagraph_source_offsets:
        #    plt.scatter(x+dx,dy)
        #plt.title(element)
        #plt.scatter(x,0, color='yellow', marker='*')
        #plt.show()
        for r_img,theta_img,dx,dy in self.coronagraph_source_offsets:

            for wave in wavelengths:
                longname = '{0:}_{1:}_{2:.4f}_{3:.3f}_{4:.3f}.fits'.format(self.insname,self.aperture_name,wave,dx,dy)
                    # Move existence check here first.
                if os.path.isfile(longname):
                    print("{0:} already exists. skipping...".format(longname))
                else:
                    psf = {'psf':self.instrument, 'wave':wave*1e-6, 'filename':longname, 'optimal_offset_x': dx, 'optimal_offset_y': dy, 'source_offset_r': r_img, 'source_offset_theta': theta_img}
                    psfs.append(psf)
        if len(psfs) > 0:
            self._split_multi(n_cores,psfs)


class niriss(JWSTInstrument):

    def __init__(self, aperture):
        """
        Set up a NIRISS instrument and aperture, and NIRISS-specific features
        """
        self.instrument = wp.NIRISS()
        self.insname = 'niriss'

        # load configuration
        super().__init__(aperture)


class nirspec(JWSTInstrument):

    def __init__(self, aperture):
        """
        Set up a NIRSpec instrument and aperture, and NIRSpec-specific features
        """
        self.instrument = wp.NIRSpec()
        self.insname = 'nirspec'

        # load configuration
        super().__init__(aperture)


class nircam(JWSTInstrument):

    def __init__(self, aperture):
        """
        Set up a NIRCam instrument and aperture, and NIRCam-specific features

        Methods
        -------
        create_psfs:
            Override the default version so that bar PSFs can be handled differently
        create_coro_psfs:
            Special routine for handling the maskswb and masklwb bar occulters
        _bar_width:
            Returns the width of the coronagraphic wedge, as a function of filter
        """
        self.instrument = wp.NIRCam()
        self.insname = 'nircam'

        # load configuration
        super().__init__(aperture)

    def create_psfs(self,n_cores):
        if self.aperture in ['masklwb', 'maskswb']:
            psfs = self.create_coro_psfs(n_cores)
        else:
            psfs = PSFInstrument.create_psfs(self,n_cores)
        return

    # this function provides the bar width as a function of X (i.e. bar_offset in arcsec).
    # the source offsets will be in units of bar width for consistency.
    def _bar_width(self,x):
        width = self.barwidth_0 + self.barwidth_1 * x
        return width

    def create_coro_psfs(self,n_cores):
        """
        Set up coronagraphy PSF sets for 2D patterns.

        For NIRCam, we set up a wedge of points that can be reflected across
        the X axis to form a complete pattern.

        The NIRCam bars are actually wedges of varying width. Given the PSF size
        at different wavelengths, there are different optimal locations for
        each filter at which the system will set the star to be occulted.

        The bar slopes are in different directions, but the offsets are
        uniformly in the direction of inreasing with longer wavelength, so we
        will need to flip the offsets for MASKLWB

        Parameters
        ----------
        n_cores: int
            The number of parallel generation instances to run
        """

        # get the optimal offset for each filter from the instrument config file.
        # the offsets are in arcseconds and increase with bar thickness (i.e.
        # negative offsets are at the narrower end; positive at wider end).
        # Also get which filters go with each bar from the config file.
        bar_offsets = self.config['bar_offsets']
        waveelements = self._get_wave_elements()
        print(waveelements)
        # We will set bar offsets, not rely on WebbPSF's default settings
        self.instrument.options['bar_offset'] = 0.0

        # the bars are rotated -90 degrees w.r.t. the webbpsf coordinate system
        bar_angle = 90.0

        # loop through the filters and make a set of PSFs for each one centered at the optimal bar position
        for element in waveelements:
            psfs = []
            # get the optimal bar position for this filter
            x = bar_offsets[element]

            self.aperture_name = "{0:}{1:}".format(self.aperture,element)
            print(element)
            wavelengths = self._get_wave_range_fixedgrid([element])

            self.coronagraph_source_offsets = []

            # now loop through the offsets and angles for positions around the optimal position
            # reverse the order so that the unocculted PSF is used to determine pupil throughput
            for dx in self.offsets_x[::-1]:
                # find bar width at this x position
                x_img = x + dx  # this is because of webbpsf X flip...
                width = self._bar_width(x_img)
                # The LWB wedge is slanted in the opposite direction from SWB.
                # Widths correctly increase for longer wavelengths, but physical
                # offsets need to be in the opposite direction. Ultimately,
                # we'd prefer to change the equation coefficients and offsets,
                # rather than have this line:
                if self.aperture == 'masklwb':
                    x_img = -1 * x_img
                    dx = -1*dx
                for dy in self.offsets_y[::-1]:


                    # compute y values for this bar width at this x position
                    dy = width * dy  # offsets in units of bar widths
                    r_img = np.sqrt(x_img**2 + dy**2)
                    theta_img = np.degrees(np.arctan2(dy, x_img)) - bar_angle
                    self.coronagraph_source_offsets.append([r_img,theta_img,dx,dy])


            ## This code can be used, when uncommented, to display the PSF locations on the obscuration element.
            #lwb = webbpsf.optics.NIRCam_BandLimitedCoron(name='MASKSWB')
            #lwb.display(grid_size=10, colorbar=False, title=element)
            #for r_img,theta_img,dx,dy in self.coronagraph_source_offsets:
            #    plt.scatter(x+dx,dy)
            #plt.title(element)
            #plt.scatter(x,0, color='yellow', marker='*')
            #plt.show()
            for r_img,theta_img,dx,dy in self.coronagraph_source_offsets:
                for wave in wavelengths:
                    longname = '{0:}_{1:}_{2:.4f}_{3:.3f}_{4:.3f}.fits'.format(self.insname,self.aperture_name,wave,dx,dy)
                    if os.path.isfile(longname):
                        print("{0:} already exists. skipping...".format(longname))
                    else:
                        psf = {'psf':self.instrument, 'wave':wave*1e-6, 'filename':longname, 'optimal_offset_x': dx, 'optimal_offset_y': dy, 'source_offset_r': r_img, 'source_offset_theta': theta_img}
                        psfs.append(psf)
            if len(psfs) > 0:
                self._split_multi(n_cores,psfs)


class wfi(RomanInstrument):

    def __init__(self, aperture):
        """
        Set up a Roman WFI instrument and aperture, and WFI-specific features
        """
        self.instrument = wp.roman.WFI()
        self.insname = "wfi"

        # load configuration
        super().__init__(aperture)


if __name__ == "__main__":
    data_dir = os.environ['pandeia_refdata']

    if len(argv) > 1:
        n_cores = int(argv[1])
    else:
        print("Usage: python make_psf.py cores instrument,aperture,... ...")
    if len(argv) > 2:
        generate = argv[2:]
    else:
        config_file = open('{0:}/{1:}/{2:}/config.json'.format(data_dir,'jwst','telescope'))
        config = json.load(config_file)
        generate = config['instruments']

    for ins in generate:
        instrument = ins.split(',')[0]
        apertures = ins.split(',')[1:]
        if len(apertures) == 0:
            config_file = open('{0:}/{1:}/{2:}/config.json'.format(data_dir,'jwst',instrument))
            config = json.load(config_file)
            apertures = config['apertures']
        if COMPRESSION:
            print("Using fpack compression, quantization level {0:}".format(COMPRESSION))
        print(instrument,apertures)
        for aperture in apertures:
            if instrument == 'miri':
                generator = miri(aperture)
            if instrument == 'nircam':
                generator = nircam(aperture)
            if instrument == 'niriss':
                generator = niriss(aperture)
            if instrument == 'nirspec':
                generator = nirspec(aperture)
            if instrument == 'wfi':
                generator = wfi(aperture)
            generator.create_psfs(n_cores)
