import sys
import os
import time
import tracemalloc

import numpy as np

from astropy import units as u
import synphot as syn
from pandeia.engine import calc_utils,perform_calculation, custom_exceptions

def analysis(outfile,filename):
    """
    Let's attack the file size problem in five ways:

    1. File size (bytes)
    2. Resolution (min/med/max)
    3. Wavelength range of file (could be truncated for filesize gains, for instance)
    Run a NIRISS WFSS calculation
    4. How long does it take? (seconds)
    5. What's the max memory usage? (bytes)

    """
    # 1: The filesize
    filesize = os.path.getsize(filename)

    try:
        spectrum = syn.SourceSpectrum.from_file(filename)
    except KeyError:
        try:
            spectrum = syn.SourceSpectrum.from_file(filename, flux_col="g00")
        except (ValueError, KeyError) as err:
            outfile.write("{},{}\n".format(filename,err))
            return
    except (TypeError, AttributeError, IndexError) as err:
        outfile.write("{},{}\n".format(filename,err))
        return
    
    # 2: Resolution
    try:
        wave = spectrum.waveset.to_value(u.micron)
        print(wave)
    except (syn.exceptions.DuplicateWavelength, TypeError, AttributeError, IndexError) as err:
        outfile.write("{},{}\n".format(filename,err))
        return
    dl = wave[1:]-wave[:-1]
    power = wave[:-1]/dl
    r_min = np.min(power)
    r_med = np.median(power)
    r_max = np.max(power)

    # 3. Wavelength range of file
    wave_min = np.min(wave)
    wave_max = np.max(wave)

    JWST="Y"
    if wave_min > 0.5 or wave_max < 31.0:
        JWST="N"
    if wave_min < 0.78 and wave_max > 1.03:

        # Run a NIRISS WFSS Imaging calculation and find out
        # 4. How much time it takes
        # 5. How much memory it takes

        calc = calc_utils.build_default_calc('jwst','niriss','wfss')
        flux = syn.units.convert_flux(spectrum.waveset,spectrum(spectrum.waveset),u.mJy).to_value(u.mJy)
        calc['scene'][0]['spectrum']['normalization']['norm_wave'] = 1.0
        calc['scene'][0]['spectrum']['normalization']['norm_flux'] = 1.0e-18
        calc['scene'][0]['spectrum']['sed'] = {'sed_type': 'input', 'spectrum' : [np.asarray(wave,dtype=np.float64),np.asarray(flux, dtype=np.float64)]}
        tracemalloc.start()
        time0 = time.time()
        try:
            result = perform_calculation.perform_calculation(calc)

            time1 = time.time()
            current,peak = tracemalloc.get_traced_memory()
            tracemalloc.stop()

            #print(f"{filename},\t{filesize} b,\t{r_min},\t{r_med},\t{r_max},\t{wave_min},\t{wave_max},\t{time1-time0} s,\t{peak}")
            outfile.write("{},{},{:.2f},{:.2f},{:.2f},{:.2f},{:.2f},{},{:.4f},{}\n".format(filename, filesize, r_min, r_med, r_max, wave_min, wave_max, JWST, time1-time0, peak))
        except (TypeError, custom_exceptions.EngineInputError) as err:
            outfile.write("{},{},{:.2f},{:.2f},{:.2f},{:.2f},{:.2f},{}, {}\n".format(filename, filesize, r_min, r_med, r_max, wave_min, wave_max, JWST, err))
    else:
        outfile.write("{},{},{:.2f},{:.2f},{:.2f},{:.2f},{:.2f},{}\n".format(filename, filesize, r_min, r_med, r_max, wave_min, wave_max, JWST))


if __name__ == "__main__":
    filepath = sys.argv[1]

    with open("spectrum_analysis.csv","w") as outfile:
        outfile.write("filename,filesize,r_min,r_med,r_max,wave_min,wave_max,JWST,time,mem\n")
        for root, dirs, files in os.walk(filepath):
            for num, filename in enumerate(files):
                if filename[-5:] == ".fits":
                    print(f"{num}/{len(files)}: {filename}")
                    analysis(outfile, os.path.join(root, filename))

        outfile.close()
