#!/usr/bin/env python
#
# This file is the source of all of the data-driven client validation rules.
# This was abstracted out of the client part of the repo, and converted to
# Python (from JS) so that it could also be used to generate client tests.
#
import os, sys

VLAD_CONSTRAINTS_JWST = {
    "niriss_soss_ord2_sub96": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #2948",
        "conditions":    [
            { "data_name":  "s_inst_mode",           "==": "niriss_soss" },
            { "element_id": "niriss_soss_subarrays", "==": "substrip96" },
            { "element_id": "strgy_g_order",         "==": "2" },
        ],
        "message":       "Order 2 not supported for subarray substrip96.",
        "msg_label_ids": ['valmsg_subarray','valmsg_strgy_g_order'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_strategy', 'btn_detectorsetup'],
        "test_values":   {'strgy_g_order': ['2','1']}, # default calc uses substrip96
        "trigger_ids":   ['strgy_g_order', 'niriss_soss_subarrays']
    },

    "niriss_soss_sub_with_nis": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #3949, bughunt-31",
        "conditions": [
            { "data_name":  "s_inst_mode",                 "==":       "niriss_soss" },
            { "element_id": "niriss_soss_subarrays",       "contains": "substrip" },
            { "element_id": "niriss_soss_readoutpatterns", "!=":       "nisrapid" },
        ],
        "message":       "When using a subarray in NIRISS SOSS, only NISRAPID readout pattern is supported.",
        "msg_label_ids": ['valmsg_readmode_3'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_detectorsetup'],
        "test_values":   {'niriss_soss_readoutpatterns': ['nis','nisrapid']}, # default calc uses substrip96
        "trigger_ids":   ['niriss_soss_subarrays','niriss_soss_readoutpatterns']
    },

    "niriss_soss_ord2_f277w": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-716",
        "conditions":    [
            { "data_name":  "s_inst_mode",           "==": "niriss_soss" },
            { "element_id": "niriss-soss-filter",    "==": "f277w" },
            { "element_id": "strgy_g_order",         "!=": "1" },
        ],
        "message":       "Only order 1 is supported for filter f277w.",
        "msg_label_ids": ['valmsg_niriss_soss_inst_setup_2','valmsg_strgy_g_order_2'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_instrumentsetup', 'btn_strategy'],
        "test_values":   {}, # Not ready for test yet, since we take out the f277w filter choice temporarily. When we are ready, the values should be 'niriss-soss-filter':  ['f277w','clear'], 'strgy_g_order': ['2','1']
        "trigger_ids":   ['niriss-soss-filter', 'strgy_g_order']
    },

    # For #3949, bughunt item no. 33 - for it we allowed nis|nisrapid, but both should ONLY be shown when subby=FULL
    # NOTE - this is due to a design inability in our pandeia_data scheme (rdmd's are not dependent on subarray)
    #        Therefore this (unlike the rest of these) is NOT easily fixed by removing a choice and adding a mig. step.
    # Text and condition changed for #4516
    "niriss_ami_sub_with_nis": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #3949, bughunt-33",
        "conditions": [
            { "data_name":  "s_inst_mode",                "==": "niriss_ami" },
            { "element_id": "niriss_ami_subarrays",       "==": "sub80" },
            { "element_id": "niriss_ami_readoutpatterns", "!=": "nisrapid" },
        ],
        "message":       "When using the SUB80 subarray in NIRISS AMI, only NISRAPID readout pattern is supported.",
        "msg_label_ids": ['valmsg_subarray_3'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  [],
        "test_values":   {"niriss_ami_readoutpatterns": ['nis','nisrapid']}, # use fact that default ami uses sub80
        "trigger_ids":   ['niriss_ami_subarrays','niriss_ami_readoutpatterns']
    },

    "nirspec_fs_irs2_w_full": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #2875",
        "conditions": [
            { "data_name":  "s_inst_mode",                        "==":       "nirspec_fixed_slit" },
            { "element_id": "nirspec_fixed_slit_readoutpatterns", "contains": "nrsirs2" },
            { "element_id": "nirspec_fixed_slit_subarrays",       "!=":       "full" },
        ],
        "message":       "IRS2 read modes may only be used if Full frame subarray is chosen.",
        "msg_label_ids": ['valmsg_subarray_2'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_detectorsetup'],
        "test_values":   {'nirspec_fixed_slit_readoutpatterns': ['nrsirs2','nrs']}, # default calc does not use full
        "trigger_ids":   ['nirspec_fixed_slit_subarrays','nirspec_fixed_slit_readoutpatterns']
    },

    # same constraint as nirspec_bots_clear_w_sub1024a, but with more to trigger it
    "nirspec_fs_clear_w_sub1024a": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "pandeia_data/pull/402",
        "conditions": [
            { "data_name":  "s_inst_mode",                  "==": "nirspec_fixed_slit" },
            { "element_id": "nirspec-fixed_slit-filtdisp",  "==": "prism,clear" },
            { "element_id": "nirspec-fixed_slit-aperture",  "==": "s1600a1" },
            { "element_id": "nirspec_fixed_slit_subarrays", "==": "sub1024a" },
        ],
        "message":       "Use of the SUB1024A subarray with the PRISM/CLEAR grating/filter pair is not supported.",
        "msg_label_ids": ['valmsg_subarray_4', 'valmsg_nirspec_fixed_slit_inst_setup_2'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_detectorsetup'],
        "test_values":   {'nirspec-fixed_slit-aperture':  ['s1600a1','s200a1'], # default calc uses "prism,clear"
                          'nirspec_fixed_slit_subarrays': ['sub1024a','full']},
        "trigger_ids":   ['nirspec-fixed_slit-filtdisp','nirspec-fixed_slit-aperture','nirspec_fixed_slit_subarrays']
    },

    # same constraint as nirspec_fs_clear_w_sub1024a, but with less to trigger it
    "nirspec_bots_clear_w_sub1024a": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-365",
        "conditions": [
            { "data_name":  "s_inst_mode",            "==": "nirspec_bots" },
            { "element_id": "nirspec-bots-filtdisp",  "==": "prism,clear" },
            { "element_id": "nirspec_bots_subarrays", "==": "sub1024a" },
        ],
        "message":       "Use of the SUB1024A subarray with the PRISM/CLEAR grating/filter pair is not supported.",
        "msg_label_ids": ['valmsg_subarray_8', 'valmsg_nirspec_bots_inst_setup_2'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_detectorsetup'],
        "test_values":   {'nirspec_bots_subarrays': ['sub1024a','sub512']}, # default calc uses "prism,clear"
        "trigger_ids":   ['nirspec-bots-filtdisp','nirspec_bots_subarrays']
    },

    "nirspec_target_acq_msata_nrsrapidd1_2_wo_clear": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-1744",
        "conditions": [
            { "data_name":  "s_inst_mode",                        "==": "nirspec_target_acq" },
            { "element_id": "nirspec-target_acq-aperture",        "==": "shutter" },
            { "element_id": "nirspec-target_acq-filter",          "!=": "clear" },
            { "element_id": "nirspec_target_acq_readoutpatterns", "contains": "nrsrapidd"},
            { "element_id": "nirspec_target_acq_readoutpatterns", "!=": "nrsrapidd6"}
        ],
        "message":       "Use of the NRSRAPIDD1/D2 readout patterns with MSATA and filters other than CLEAR are not supported.",
        "msg_label_ids": ['valmsg_readmode_4'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_detectorsetup'],
        "test_values":   {'nirspec-target_acq-aperture': ['shutter','s1600a1'],
                          'nirspec_target_acq_readoutpatterns': ['nrsrapidd1','nrsrapidd6']},
        "trigger_ids":   ['nirspec-target_acq-aperture','nirspec-target_acq-filter','nirspec_target_acq_readoutpatterns']
    },

    "nircam_coro_w_filt_plus_nd": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #3949, bughunt-21",
        "conditions": [
            { "data_name":  "s_inst_mode",                "==":       "nircam_coronagraphy" },
            { "element_id": "nircam-coronagraphy-filter", "contains": "_nd" },
        ],
        "message":       "Filter+ND option is only supported in Target Acquisition mode.",
        "msg_label_ids": ['valmsg_nircam_coronagraphy_inst_setup'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  ['btn_instrumentsetup'],
        "test_values":   {"nircam-coronagraphy-filter": ["f210m_nd","f210m"]},
        "trigger_ids":   ['nircam-coronagraphy-filter']
    },

    "nircam_sw_ts_lw_ts_warn_docs": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-347",
        "conditions": [
            { "data_name":  "s_inst_mode",          "==": "nircam_sw_ts" },
            { "element_id": "nircam_sw_ts_submode", "==": "lw_ts" },
        ],
        "message":       'The Pupil chosen in this tab limits the options you can use in the simultaneous LW Time Series observation in APT. <a class="etc_help_a" id="help_wkbookpg_calcpage_insttab_sw_ts_msg"><span class="glyphicon glyphicon-question-sign"></span></a>',
        "msg_label_ids": ['valmsg_nircam_sw_ts_inst_setup'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  [],
        "test_values":   {"nircam_sw_ts_submode": ["lw_ts","ssgrism"], "default_calc_triggers": 1},
        "trigger_ids":   ['nircam_sw_ts_submode']
    },

    "nircam_sw_ts_lw_ts_wlp8_sub64": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-262",
        "conditions": [
            { "data_name":  "s_inst_mode",            "==":       "nircam_sw_ts" },
            { "element_id": "nircam_sw_ts_submode" ,  "==":       "lw_ts" }, # in default calc
            { "element_id": "nircam-sw_ts-aperture",  "contains": "wlp8" },
            { "element_id": "nircam_sw_ts_subarrays", "contains": "64" },
        ],
        "message":       "The WLP8 PSF is significantly larger than this subarray, so photometric precision will be compromised. SUB160P or larger is recommended.",
        "msg_label_ids": ['valmsg_subarray_5'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  ['btn_instrumentsetup','btn_detectorsetup'],
        "test_values":   {'nircam-sw_ts-aperture': ['wlp8__ts','sw'],
                          'nircam_sw_ts_subarrays': ['sub64p','full']},
        "trigger_ids":   ['nircam-sw_ts-aperture','nircam_sw_ts_subarrays']
    },

    "nircam_sw_ts_ssgrism_wlp8_sub64": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-262",
        "conditions": [
            { "data_name":  "s_inst_mode",            "==":       "nircam_sw_ts" },
            { "element_id": "nircam_sw_ts_submode" ,  "==":       "ssgrism" },
            { "element_id": "nircam-sw_ts-aperture",  "contains": "wlp8" },
            { "element_id": "nircam_sw_ts_subarrays", "contains": "64" },
        ],
        "message":       "The WLP8 PSF is significantly larger than this subarray, so photometric precision will be compromised. SUBGRISM128 or larger is recommended.",
        "msg_label_ids": ['valmsg_subarray_6'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  ['btn_instrumentsetup','btn_detectorsetup'],
        "test_values":   {'nircam_sw_ts_submode': ['ssgrism','lw_ts'],
                          'nircam-sw_ts-aperture': ['wlp8__tsgrism','sw'],
                          'nircam_sw_ts_subarrays': ['subgrism64','full']},
        "trigger_ids":   ['nircam-sw_ts-aperture','nircam_sw_ts_subarrays']
    },

    # s_disperser being a data_name, not an element_id, forces the vlad to mean post-calc
    "nircam_wfgrism_grismc_orientation": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #3963",
        "conditions": [
            { "data_name": "s_inst_mode",  "==": "nircam_wfgrism" },
            { "data_name": "s_disperser",  "==": "grismc" },
        ],
        "message":       "Note that in the NIRCam Wide Field Grism mode, the scene is correctly dispersed in the row or column direction. For GRISMC, the dispersed 2D plots are rotated 90 deg clockwise.",
        "msg_label_ids": ['hdr_msg_2dplots'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  [],
        "test_values":   {}, # not yet easy to test
        "trigger_ids":   ["calc_pane_msg"]
    },

    "miri_imaging_fnd_wout_full": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-489",
        "conditions": [
            { "data_name":  "s_inst_mode",            "==":  "miri_imaging" },
            { "element_id": "miri-imaging-filter",    "==":  "fnd" },
            { "element_id": "miri_imaging_subarrays", "!=":  "full" },
        ],
        "message":       "The neutral density filter is not available for science applications and must be used with the FULL subarray.",
        "msg_label_ids": ['valmsg_subarray_7'],
        "severity":      "pandeia.const.ERROR",
        "tab_link_ids":  ['btn_detectorsetup'],
        "test_values":   {'miri-imaging-filter': ["fnd","f560w"],
                          'miri_imaging_subarrays': ['sub64','full']},
        "trigger_ids":   ['miri-imaging-filter','miri_imaging_subarrays']
    },

    "miri_modes_w_unsupported_slow_bh7": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #3949, bughunt-7",
        "conditions": [
            { "data_name":  "s_inst_mode",                       "==":  "miri_lrsslitless" },
            { "element_id": "miri_lrsslitless_readoutpatterns",  "==":  "slowr1" },
        ],
        "message":       "The SLOWR1 readout pattern is not supported for this mode.",
        "msg_label_ids": ['valmsg_readmode'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  [],
        "test_values":   {"miri_lrsslitless_readoutpatterns": ["slowr1","fastr1"]},
        "trigger_ids":   ['miri_lrsslitless_readoutpatterns']
    },

    "miri_modes_w_unsupported_slow_bh8": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "github #3949, bughunt-8",
        "conditions": [
            { "data_name":  "s_inst_mode",                        "==":  "miri_coronagraphy" },
            { "element_id": "miri_coronagraphy_readoutpatterns",  "==":  "slowr1" },
        ],
        "message":       "The SLOWR1 readout pattern is not supported for this mode.",
        "msg_label_ids": ['valmsg_readmode_2'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  [],
        "test_values":   {"miri_coronagraphy_readoutpatterns": ["slowr1","fastr1"]},
        "trigger_ids":   ['miri_coronagraphy_readoutpatterns']
    },

    # Must not give tab link for this one, would counteract numeric input box checker.
    # Use >= 2 since 2 is the inclusive min-allowed val in numeric input box.
    "miri_low_ngroups_vals": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "want miri ngroups > 5",
        "conditions": [
            { "data_name":  "s_instrument",  "==": "miri" },
            { "data_name":  "s_mode",        "!=": "target_acq" },
            { "element_id": "ngroups_input", ">=": 2 },
            { "element_id": "ngroups_input", "<":  5 },
        ],
        "message":       "NGROUPS >= 5 are strongly recommended. NGROUPS of 2 are a limited access option, and NGROUPS of 3 or 4 are allowed but discouraged (see MIRI documentation for more details).",
        "msg_label_ids": ['valmsg_ngroups'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  [],
        "test_values":   {"ngroups_input": [3,6]},
        "trigger_ids":   ["ngroups_input"]
    },

    "miri_time_single_int": {
        "area_name":     "pandeia.const.AREA_CALC",
        "comment":       "JETC-301",
        "conditions": [
            { "data_name":  "s_instrument",    "==": "miri" },
            { "data_name":  "s_mode",          "!=": "target_acq" },
            { "element_id": "calc_single_int", ">": 2000 }
        ],
        "message":       "Individual integrations >2000 seconds are not allowed for this mode in APT.",
        "msg_label_ids": ['valmsg_ngroups_2'],
        "severity":      "pandeia.const.WARNING",
        "tab_link_ids":  [],
        "test_values":   {"ngroups_input": [10000,100]},
        "trigger_ids":   ["calc_single_int"]  # may later change calc_single_int to work via a ${var1} tmp var
    },
}

VLAD_CONSTRAINTS_HST = {
    'stis_minimum_time': {
        'area_name':     'pandeia.const.AREA_CALC',
        'comment':       'JETC-1951',
        'conditions': [
            {'data_name':  's_inst_mode',           '==': 'stis_ccd_imaging'},
            {'let':        'var1',                  'element_id': 'stis_ccd_imaging_nsplit', '*=': 0.0999999999999},
            {'element_id': 'stis_ccd_imaging_time', '<': '${var1}'}
        ],
        'message':       'Time specified must be no less than 0.1 s per frame.',
        'msg_label_ids': ['valmsg_time_1'],
        'severity':      'pandeia.const.ERROR',
        'tab_link_ids':  ['btn_detectorsetup'],
        'test_values':   {'stis_ccd_imaging_time': [0.0001,10]},
        'trigger_ids':   ['stis_ccd_imaging_time','stis_ccd_imaging_nsplit']
    },
}

# Dev notes:
#
# A now-removed vlad named "miri_slow_nints_gt_1" used two features now unused:
#   can put "__s_mode__" into element_id value to be evaluated in the client.
#   can put "${ALL_INST_MODES}" into trigger_ids values to have it expanded during the test regen
